
/*******************************************************************************
*
* File plaq2.c
*
* Copyright (C) 2015 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Exact leading-order contribution to the plaquette loops.
*
*   void plaq2(double **plaq)
*     Computes the leading-order contribution to the plaquette loop in
*     all 6 planes as a function of the time x0.
*
* To leading order in the gauge coupling, the plaquette loop in the (mu,nu)
* plane is given by
*
*   (1/4)*<F^a_{mu,nu}(x)F^a_{mu,nu}(x)>,
*
*   F_{mu,nu}(x)=A_nu(x+mu)-A_nu(x)-A_mu(x+nu)+A_mu(x).
*
* The program plaq2() computes the expectation value exactly and assigns the
* calculated value to
*
*   plaq[n][x0], n=0,..,5, x0=0,..,T-1,
*
* where n counts the planes (mu,nu)=[(0,1),(0,2),(0,3),(1,2),(1,3),(2,3)] and
* T=NPROC0*L0.
*
* The program plaq2() acts globally and must be called on all MPI processes
* simultaneously.
*
*******************************************************************************/

#define PLAQ2_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "utils.h"
#include "pmdb.h"
#include "global.h"

#define N0 (NPROC0*L0)
#define N1 (NPROC1*L1)
#define N2 (NPROC2*L2)
#define N3 (NPROC3*L3)

static const int plns[6][2]={{0,1},{0,2},{0,3},{2,3},{3,1},{1,2}};
static int init=0;
static double *phat[4],co[L0][N0],si[L0][N0],loc_plaq[6][N0];


static void set_phat(void)
{
   int l[4],mu,k,k0,x0,bc;
   double r[4],pi,p0;

   phat[0]=malloc((L0+L1+L2+L3)*sizeof(double));
   error(phat[0]==NULL,1,"set_phat [plaq2.c]",
         "Unable to allocate auxiliarly array");
   phat[1]=phat[0]+L0;
   phat[2]=phat[1]+L1;
   phat[3]=phat[2]+L2;

   pi=4.0*atan(1.0);
   r[0]=pi/(double)(2*N0);
   r[1]=pi/(double)(N1);
   r[2]=pi/(double)(N2);
   r[3]=pi/(double)(N3);

   l[0]=L0;
   l[1]=L1;
   l[2]=L2;
   l[3]=L3;

   bc=bc_type();

   for (mu=0;mu<4;mu++)
   {
      k0=cpr[mu]*l[mu];

      if ((bc==1)||(mu>0))
      {
         for (k=0;k<l[mu];k++)
            phat[mu][k]=2.0*sin((double)(k+k0)*r[mu]);
      }
      else
      {
         for (k=0;k<l[mu];k++)
            phat[mu][k]=2.0*sin(((double)(k+k0)+0.5)*r[mu]);
      }
   }

   k0=cpr[0]*L0;

   for (k=0;k<L0;k++)
   {
      if (bc==1)
      {
         p0=(double)(k+k0)*pi/(double)(N0);

         for (x0=0;x0<N0;x0++)
         {
            co[k][x0]=cos(p0*((double)(x0)+0.5));
            si[k][x0]=sin(p0*(double)(x0));
         }
      }
      else
      {
         p0=((double)(k+k0)+0.5)*pi/(double)(N0);

         for (x0=0;x0<N0;x0++)
         {
            co[k][x0]=cos(p0*(double)(x0));
            si[k][x0]=sin(p0*((double)(x0)+0.5));
         }
      }
   }

   init=1;
}


void plaq2(double **plaq)
{
   int bc,n,x0,mu,nu;
   int k0,k1,k2,k3;
   double ph[4],phsq,r0,r1;

   if (init==0)
      set_phat();

   bc=bc_type();

   for (n=0;n<6;n++)
   {
      for (x0=0;x0<N0;x0++)
         loc_plaq[n][x0]=0.0;
   }

   for (k0=0;k0<L0;k0++)
   {
      for (k1=0;k1<L1;k1++)
      {
         for (k2=0;k2<L2;k2++)
         {
            for (k3=0;k3<L3;k3++)
            {
               if ((bc!=1)||(k0>0)||(k1>0)||(k2>0)||(k3>0)||
                   (cpr[0]>0)||(cpr[1]>0)||(cpr[2]>0)||(cpr[3]>0))
               {
                  ph[0]=phat[0][k0];
                  ph[1]=phat[1][k1];
                  ph[2]=phat[2][k2];
                  ph[3]=phat[3][k3];
                  phsq=ph[0]*ph[0]+ph[1]*ph[1]+ph[2]*ph[2]+ph[3]*ph[3];

                  for (n=0;n<6;n++)
                  {
                     mu=plns[n][0];
                     nu=plns[n][1];
                     r0=(ph[mu]*ph[mu]+ph[nu]*ph[nu])/phsq;

                     if ((k0==0)&&(bc==1)&&(cpr[0]==0))
                        r0*=0.5;

                     for (x0=0;x0<N0;x0++)
                     {
                        if (((bc==1)&&(n<3))||((bc==2)&&(n>=3)))
                           r1=co[k0][x0];
                        else
                           r1=si[k0][x0];

                        loc_plaq[n][x0]+=r0*r1*r1;
                     }
                  }
               }
            }
         }
      }
   }

   r0=4.0/((double)(N0)*(double)(N1*N2*N3));

   for (n=0;n<6;n++)
   {
      for (x0=0;x0<N0;x0++)
         loc_plaq[n][x0]*=r0;

      MPI_Reduce(loc_plaq[n],plaq[n],N0,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);
      MPI_Bcast(plaq[n],N0,MPI_DOUBLE,0,MPI_COMM_WORLD);
   }
}
