
/*******************************************************************************
*
* File cnrm1.c
*
* Copyright (C) 2015 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Calculation of the time-slice norms of the ghost field.
*
* The externally accessible functions are
*
*   void cnrm1(sl3_alg_dble *cd1,sl3_alg_dble *cd2,complex_dble *cnrm)
*     Computes the leading-order contribution to the time-slice norms
*     as a function of the time x0 (see the notes).
*
* Notes:
*
* For both SF and open-SF boundary conditions, the values assigned to
* cnrm[x0] are given by
*
*   c^a(x)*cbar^a(x) for x0=0,..,T-1,
*
* where T=NPROC0*L0 and c(x) and cbar(x) are is the ghost fields defined by
* the field arrays cd1 and cd2. The values are averaged over the space
* coordinates.
*
* The program cnrm1() acts globally and must be called on all MPI processes
* simultaneously.
*
*******************************************************************************/

#define CNRM1_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "sl3.h"
#include "utils.h"
#include "pmdb.h"
#include "lattice.h"
#include "aflds.h"
#include "global.h"

#define N0 (NPROC0*L0)
#define N1 (NPROC1*L1)
#define N2 (NPROC2*L2)
#define N3 (NPROC3*L3)

static complex_dble loc_cnrm[N0];


static complex_dble biform(sl3_alg_dble *X,sl3_alg_dble *Y)
{
   complex_dble zx,zy,sm;

   zx.re=(*X).c1.re+(*X).c2.re;
   zx.im=(*X).c1.im+(*X).c2.im;

   zy.re=(*Y).c1.re+(*Y).c2.re;
   zy.im=(*Y).c1.im+(*Y).c2.im;

   sm.re=zx.re*zy.re-zx.im*zy.im;
   sm.im=zx.re*zy.im+zx.im*zy.re;

   sm.re+=(*X).c1.re*(*Y).c1.re-(*X).c1.im*(*Y).c1.im;
   sm.im+=(*X).c1.re*(*Y).c1.im+(*X).c1.im*(*Y).c1.re;

   sm.re+=(*X).c2.re*(*Y).c2.re-(*X).c2.im*(*Y).c2.im;
   sm.im+=(*X).c2.re*(*Y).c2.im+(*X).c2.im*(*Y).c2.re;

   sm.re+=(*X).c3.re*(*Y).c4.re-(*X).c3.im*(*Y).c4.im;
   sm.im+=(*X).c3.re*(*Y).c4.im+(*X).c3.im*(*Y).c4.re;

   sm.re+=(*X).c4.re*(*Y).c3.re-(*X).c4.im*(*Y).c3.im;
   sm.im+=(*X).c4.re*(*Y).c3.im+(*X).c4.im*(*Y).c3.re;

   sm.re+=(*X).c5.re*(*Y).c6.re-(*X).c5.im*(*Y).c6.im;
   sm.im+=(*X).c5.re*(*Y).c6.im+(*X).c5.im*(*Y).c6.re;

   sm.re+=(*X).c6.re*(*Y).c5.re-(*X).c6.im*(*Y).c5.im;
   sm.im+=(*X).c6.re*(*Y).c5.im+(*X).c6.im*(*Y).c5.re;

   sm.re+=(*X).c7.re*(*Y).c8.re-(*X).c7.im*(*Y).c8.im;
   sm.im+=(*X).c7.re*(*Y).c8.im+(*X).c7.im*(*Y).c8.re;

   sm.re+=(*X).c8.re*(*Y).c7.re-(*X).c8.im*(*Y).c7.im;
   sm.im+=(*X).c8.re*(*Y).c7.im+(*X).c8.im*(*Y).c7.re;

   sm.re*=(-2.0);
   sm.im*=(-2.0);

   return sm;
}


static void local_cnrm(sl3_alg_dble *cd1,sl3_alg_dble *cd2)
{
   int ix,t;
   complex_dble w;

   for (t=0;t<N0;t++)
   {
      loc_cnrm[t].re=0.0;
      loc_cnrm[t].im=0.0;
   }

   for (ix=0;ix<VOLUME;ix++)
   {
      t=global_time(ix);
      w=biform(cd1,cd2);
      loc_cnrm[t].re+=w.re;
      loc_cnrm[t].im+=w.im;

      cd1+=1;
      cd2+=1;
   }
}


void cnrm1(sl3_alg_dble *cd1,sl3_alg_dble *cd2,complex_dble *cnrm)
{
   int t;
   double r0;

   local_cnrm(cd1,cd2);
   r0=1.0/((double)(N1*N2*N3));

   for (t=0;t<N0;t++)
   {
      loc_cnrm[t].re*=r0;
      loc_cnrm[t].im*=r0;
   }

   MPI_Reduce(loc_cnrm,cnrm,2*N0,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);
   MPI_Bcast(cnrm,2*N0,MPI_DOUBLE,0,MPI_COMM_WORLD);
}
