
/*******************************************************************************
*
* File star1.c
*
* Copyright (C) 2015 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Calculation of the "star" expectation values of the ghost field.
*
* The externally accessible functions are
*
*   void star1(sl3_alg_dble *cd1,sl3_alg_dble *cd2,complex_dble **star)
*     Computes the "star" expectation values of the ghost field as a
*     a function of the time x0 (see the notes).
*
* Notes:
*
* For both SF and open-SF boundary conditions, the values assigned to
* star[mu][x0] are given by
*
*   drv_mu c^a(x)*drv_mu cbar^a(x)  (no sum over mu),
*
* where T=NPROC0*L0 and c(x) and cbar(x) are the ghost fields defined by
* the field arrays cd1 and cd2. The values are averaged over the space
* coordinates.
*
* The program star1() acts globally and must be called on all MPI processes
* simultaneously.
*
*******************************************************************************/

#define STAR1_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "sl3.h"
#include "utils.h"
#include "pmdb.h"
#include "lattice.h"
#include "cflds.h"
#include "global.h"

#define N0 (NPROC0*L0)
#define N1 (NPROC1*L1)
#define N2 (NPROC2*L2)
#define N3 (NPROC3*L3)

static complex_dble loc_star[4][N0];


static complex_dble biform(sl3_alg_dble *X,sl3_alg_dble *Y)
{
   complex_dble zx,zy,sm;

   zx.re=(*X).c1.re+(*X).c2.re;
   zx.im=(*X).c1.im+(*X).c2.im;

   zy.re=(*Y).c1.re+(*Y).c2.re;
   zy.im=(*Y).c1.im+(*Y).c2.im;

   sm.re=zx.re*zy.re-zx.im*zy.im;
   sm.im=zx.re*zy.im+zx.im*zy.re;

   sm.re+=(*X).c1.re*(*Y).c1.re-(*X).c1.im*(*Y).c1.im;
   sm.im+=(*X).c1.re*(*Y).c1.im+(*X).c1.im*(*Y).c1.re;

   sm.re+=(*X).c2.re*(*Y).c2.re-(*X).c2.im*(*Y).c2.im;
   sm.im+=(*X).c2.re*(*Y).c2.im+(*X).c2.im*(*Y).c2.re;

   sm.re+=(*X).c3.re*(*Y).c4.re-(*X).c3.im*(*Y).c4.im;
   sm.im+=(*X).c3.re*(*Y).c4.im+(*X).c3.im*(*Y).c4.re;

   sm.re+=(*X).c4.re*(*Y).c3.re-(*X).c4.im*(*Y).c3.im;
   sm.im+=(*X).c4.re*(*Y).c3.im+(*X).c4.im*(*Y).c3.re;

   sm.re+=(*X).c5.re*(*Y).c6.re-(*X).c5.im*(*Y).c6.im;
   sm.im+=(*X).c5.re*(*Y).c6.im+(*X).c5.im*(*Y).c6.re;

   sm.re+=(*X).c6.re*(*Y).c5.re-(*X).c6.im*(*Y).c5.im;
   sm.im+=(*X).c6.re*(*Y).c5.im+(*X).c6.im*(*Y).c5.re;

   sm.re+=(*X).c7.re*(*Y).c8.re-(*X).c7.im*(*Y).c8.im;
   sm.im+=(*X).c7.re*(*Y).c8.im+(*X).c7.im*(*Y).c8.re;

   sm.re+=(*X).c8.re*(*Y).c7.re-(*X).c8.im*(*Y).c7.im;
   sm.im+=(*X).c8.re*(*Y).c7.im+(*X).c8.im*(*Y).c7.re;

   sm.re*=(-2.0);
   sm.im*=(-2.0);

   return sm;
}


static void grad_cd(sl3_alg_dble *cd,sl3_alg_dble *ad)
{
   int ix,iy,t,mu;

   cpcd_int_bnd(0,cd);

   for (ix=(VOLUME/2);ix<VOLUME;ix++)
   {
      t=global_time(ix);

      if (t==0)
      {
         iy=iup[ix][0];
         (*ad)=cd[iy];
         _sl3_alg_sub_assign(*ad,cd[ix]);
         ad+=1;

         iy=idn[ix][0];
         _sl3_alg_mulr_assign(*ad,-1.0,cd[iy]);
         ad+=1;
      }
      else if (t==(N0-1))
      {
         _sl3_alg_mulr_assign(*ad,-1.0,cd[ix]);
         ad+=1;

         iy=idn[ix][0];
         (*ad)=cd[ix];
         _sl3_alg_sub_assign(*ad,cd[iy]);
         ad+=1;
      }
      else
      {
         iy=iup[ix][0];
         (*ad)=cd[iy];
         _sl3_alg_sub_assign(*ad,cd[ix]);
         ad+=1;

         iy=idn[ix][0];
         (*ad)=cd[ix];
         _sl3_alg_sub_assign(*ad,cd[iy]);
         ad+=1;
      }

      for (mu=1;mu<4;mu++)
      {
         iy=iup[ix][mu];
         (*ad)=cd[iy];
         _sl3_alg_sub_assign(*ad,cd[ix]);
         ad+=1;

         iy=idn[ix][mu];
         (*ad)=cd[ix];
         _sl3_alg_sub_assign(*ad,cd[iy]);
         ad+=1;
      }
   }
}


static void local_star(sl3_alg_dble *ad1,sl3_alg_dble *ad2)
{
   int ix,t,mu;
   complex_dble w;

   for (mu=0;mu<4;mu++)
   {
      for (t=0;t<N0;t++)
      {
         loc_star[mu][t].re=0.0;
         loc_star[mu][t].im=0.0;
      }
   }

   for (ix=(VOLUME/2);ix<VOLUME;ix++)
   {
      t=global_time(ix);

      for (mu=0;mu<4;mu++)
      {
         w=biform(ad1,ad2);
         loc_star[mu][t].re+=w.re;
         loc_star[mu][t].im+=w.im;
         ad1+=1;
         ad2+=1;

         w=biform(ad1,ad2);

         if (mu==0)
         {
            if (t>0)
            {
               loc_star[mu][t-1].re+=w.re;
               loc_star[mu][t-1].im+=w.im;
            }
            else
            {
               loc_star[mu][N0-1].re+=w.re;
               loc_star[mu][N0-1].im+=w.im;
            }
         }
         else
         {
            loc_star[mu][t].re+=w.re;
            loc_star[mu][t].im+=w.im;
         }

         ad1+=1;
         ad2+=1;
      }
   }
}


void star1(sl3_alg_dble *cd1,sl3_alg_dble *cd2,complex_dble **star)
{
   int mu,t;
   double r;
   sl3_alg_dble **wad;

   wad=reserve_wad(2);
   grad_cd(cd1,wad[0]);
   grad_cd(cd2,wad[1]);
   local_star(wad[0],wad[1]);
   r=1.0/(double)(N1*N2*N3);

   for (mu=0;mu<4;mu++)
   {
      for (t=0;t<N0;t++)
      {
         loc_star[mu][t].re*=r;
         loc_star[mu][t].im*=r;
      }

      MPI_Reduce(loc_star[mu],star[mu],2*N0,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);
      MPI_Bcast(star[mu],2*N0,MPI_DOUBLE,0,MPI_COMM_WORLD);
   }

   release_wad();
}
