
/*******************************************************************************
*
* File time1.c
*
* Copyright (C) 2015 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Timing of the program cprop().
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "mpi.h"
#include "random.h"
#include "pmdb.h"
#include "utils.h"
#include "lattice.h"
#include "linalg.h"
#include "cflds.h"
#include "global.h"


static void new_fld(sl3_alg_dble *cd)
{
   gauss_dble((double*)(cd),16*VOLUME);
}


static double wt_cprop(void)
{
   int my_rank,nflds;
   int n,nmax,i,ib;
   double wt1,wt2,wdt,wtav;
   sl3_alg_dble **wcd;

   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   nflds=(int)((8*1024*1024)/(16*VOLUME*sizeof(double)))+1;
   nflds*=2;
   alloc_wcd(nflds);
   wcd=reserve_wcd(nflds);

   for (i=0;i<nflds;i++)
   {
      new_fld(wcd[i]);
      set_cdbc(wcd[i]);
   }

   cprop(wcd[0],wcd[1]);
   nmax=1;

   for (ib=0;ib<1;nmax*=2)
   {
      MPI_Barrier(MPI_COMM_WORLD);
      wt1=MPI_Wtime();

      for (n=0;n<nmax;n++)
      {
         for (i=0;i<nflds;i+=2)
            cprop(wcd[i],wcd[i+1]);
      }

      wt2=MPI_Wtime();
      wdt=wt2-wt1;

      MPI_Reduce(&wdt,&wtav,1,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);

      if (my_rank==0)
      {
         wtav/=(double)(NPROC);

         if (wtav>2.0)
            ib=1;

         wtav/=(double)((nmax*nflds)/2);
      }

      MPI_Bcast(&ib,1,MPI_INT,0,MPI_COMM_WORLD);
   }

   MPI_Bcast(&wtav,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   release_wcd();

   return wtav;
}


int main(int argc,char *argv[])
{
   int my_rank,bc;
   double wdt;
   FILE *flog=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("time1.log","w",stdout);

      printf("\n");
      printf("Timing of the program cprop()\n");
      printf("-----------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      bc=find_opt(argc,argv,"-bc");

      if (bc!=0)
         error_root(sscanf(argv[bc+1],"%d",&bc)!=1,1,"main [time1.c]",
                    "Syntax: time1 [-bc <type>]");

      if (NPROC>1)
         printf("There are %d MPI processes\n",NPROC);
      else
         printf("There is 1 MPI process\n");

      if ((16*VOLUME*sizeof(double))<(4096*1024))
         printf("The local size of a complex ghost field is %d KB\n",
                (int)((16*VOLUME*sizeof(double))/(1024)));
      else
         printf("The local size of a complex ghost field is %d MB\n",
                (int)((16*VOLUME*sizeof(double))/(1024*1024)));

#if (defined x64)
#if (defined AVX)
      printf("Using AVX instructions\n");
#else
      printf("Using SSE3 instructions and 16 xmm registers\n");
#endif
#if (defined P3)
      printf("Assuming SSE prefetch instructions fetch 32 bytes\n");
#elif (defined PM)
      printf("Assuming SSE prefetch instructions fetch 64 bytes\n");
#elif (defined P4)
      printf("Assuming SSE prefetch instructions fetch 128 bytes\n");
#else
      printf("SSE prefetch instructions are not used\n");
#endif
#endif
      printf("\n");
      fflush(flog);
   }

   MPI_Bcast(&bc,1,MPI_INT,0,MPI_COMM_WORLD);
   set_th_parms(LYM,bc,0);
   print_th_parms();
   start_ranlux(0,12345);
   geometry();
   wdt=wt_cprop();

   if (my_rank==0)
   {
      printf("Time per application of cprop(): %.1e sec\n",wdt);
      printf("Time per lattice point: %4.3f usec\n\n",
             1.0e6*wdt/(double)(VOLUME));
   }

   if (my_rank==0)
      fclose(flog);

   MPI_Finalize();
   exit(0);
}
