
/*******************************************************************************
*
* File check1.c
*
* Copyright (C) 2015, 2017 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Check of the programs Eclov() and Sclov().
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <float.h>
#include "mpi.h"
#include "sl3.h"
#include "random.h"
#include "pmdb.h"
#include "utils.h"
#include "lattice.h"
#include "su3fcts.h"
#include "linalg.h"
#include "aflds.h"
#include "uflds.h"
#include "ftensor.h"
#include "global.h"

#define NMAX 4

static su3_dble vd ALIGNED16;
static su3_dble X ALIGNED16;
static su3_dble Y ALIGNED16;
static complex_dble E1_sp[NMAX+1],E1_tm[NMAX+1];
static complex_dble E2_sp[NMAX+1],E2_tm[NMAX+1];
static complex_dble S1[NMAX+1],S2[NMAX+1];


static void new_fld(int n,sl3_alg_dble **ad)
{
   int i;

   for (i=0;i<=n;i++)
   {
      if (cpr[0]==(NPROC0-1))
         gauss_dble((double*)(ad[i]),16*(4*VOLUME+7*(BNDRY/4)+3));
      else
         gauss_dble((double*)(ad[i]),16*(4*VOLUME+7*(BNDRY/4)));

      set_adbc(ad[i]);
   }
}


static void global_transform(int n,sl3_alg_dble **ad)
{
   int i,ix,ifc;
   sl3_alg_dble *fld;

   for (i=0;i<=n;i++)
   {
      fld=ad[i];

      for (ix=(VOLUME/2);ix<VOLUME;ix++)
      {
         for (ifc=0;ifc<8;ifc++)
         {
            sl3_to_su3(fld,&X);
            su3xsu3(&vd,&X,&Y);
            su3xsu3dag(&Y,&vd,&X);
            su3_to_sl3(&X,fld);
            fld+=1;
         }
      }
   }
}


static void random_vec(int *svec)
{
   int mu,bs[4];
   double r[4];

   bs[0]=NPROC0*L0;
   bs[1]=NPROC1*L1;
   bs[2]=NPROC2*L2;
   bs[3]=NPROC3*L3;
   svec[0]=0;
   svec[1]=0;
   svec[2]=0;
   svec[3]=0;

   while ((abs(svec[1])+abs(svec[2])+abs(svec[3]))==0)
   {
      ranlxd(r,4);

      for (mu=1;mu<4;mu++)
      {
         svec[mu]=(int)((double)(bs[mu])*r[mu]);
         if (svec[mu]>(bs[mu]/2))
            svec[mu]-=bs[mu];
      }
   }

   MPI_Bcast(svec,4,MPI_INT,0,MPI_COMM_WORLD);
}


static void shift_flds(int n,int *s,sl3_alg_dble **ad)
{
   int i;

   for (i=0;i<=n;i++)
      shift_ad(s,ad[i]);
}


static void devES(int n,double *dev)
{
   int i;
   double d,nrm[2];
   complex_dble z;

   dev[0]=0.0;
   dev[1]=0.0;
   nrm[0]=0.0;
   nrm[1]=0.0;

   for (i=0;i<=n;i++)
   {
      z.re=E1_sp[i].re-E2_sp[i].re;
      z.im=E1_sp[i].im-E2_sp[i].im;
      d=z.re*z.re+z.im*z.im;

      if (d>dev[0])
         dev[0]=d;

      z.re=E1_tm[i].re-E2_tm[i].re;
      z.im=E1_tm[i].im-E2_tm[i].im;
      d=z.re*z.re+z.im*z.im;

      if (d>dev[0])
         dev[0]=d;

      nrm[0]+=E1_sp[i].re*E1_sp[i].re+E1_sp[i].im*E1_sp[i].im;

      z.re=S1[i].re-S2[i].re;
      z.im=S1[i].im-S2[i].im;
      d=z.re*z.re+z.im*z.im;

      if (d>dev[1])
         dev[1]=d;

      nrm[1]+=S1[i].re*S1[i].re+S1[i].im*S1[i].im;
   }

   nrm[0]/=(double)(n);
   nrm[1]/=(double)(n);

   dev[0]=sqrt(dev[0]/nrm[0]);
   dev[1]=sqrt(dev[1]/nrm[1]);
}


int main(int argc,char *argv[])
{
   int my_rank,bc;
   int n,s[4];
   double dev[2];
   sl3_alg_dble **wad;
   su3_dble **wud;
   FILE *flog=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);
   set_error_file("check1.log",0);

   if (my_rank==0)
   {
      flog=freopen("check1.log","w",stdout);

      printf("\n");
      printf("Check of the programs Eclov() and Sclov()\n");
      printf("-----------------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      bc=find_opt(argc,argv,"-bc");
      error_root((bc==0)||(sscanf(argv[bc+1],"%d",&bc)!=1),1,
                 "main [check1.c]","Syntax: check1 -bc <type>");
      fflush(flog);
   }

   MPI_Bcast(&bc,1,MPI_INT,0,MPI_COMM_WORLD);
   set_th_parms(LYM,bc,0);
   print_th_parms();
   start_ranlux(0,12345);
   geometry();
   alloc_wad(NMAX+1);
   alloc_wud(NMAX+1);
   wad=reserve_wad(NMAX+1);
   wud=reserve_wud(NMAX+1);

   if (my_rank==0)
      printf("Global gauge invariance:\n");

   for (n=0;n<=NMAX;n++)
   {
      random_su3_dble(&vd);
      MPI_Bcast(&vd,18,MPI_DOUBLE,0,MPI_COMM_WORLD);
      new_fld(n,wad);

      a2uflds(n,wad,wud[0]);
      Eclov(n,wud[0],E1_sp,E1_tm);
      Sclov(n,wud[0],S1);

      global_transform(n,wad);
      a2uflds(n,wad,wud[0]);
      Eclov(n,wud[0],E2_sp,E2_tm);
      Sclov(n,wud[0],S2);

      devES(n,dev);

      if (my_rank==0)
      {
         printf("n=%d, dev(E) = %.1e, dev(S) = %.1e\n",n,dev[0],dev[1]);
         fflush(flog);
      }
   }

   if (my_rank==0)
   {
      printf("\n");
      printf("Translation invariance:\n");
   }

   for (n=0;n<NMAX;n++)
   {
      new_fld(n+1,wad);

      a2uflds(n,wad,wud[0]);
      Eclov(n,wud[0],E1_sp,E1_tm);
      Sclov(n,wud[0],S1);

      random_vec(s);
      shift_flds(n,s,wad);

      a2uflds(n,wad,wud[0]);
      Eclov(n,wud[0],E2_sp,E2_tm);
      Sclov(n,wud[0],S2);

      devES(n,dev);

      if (my_rank==0)
      {
         printf("n=%d, s= % 3d,% 3d,% 3d,% 3d, dev(E) = %.1e, dev(S) = %.1e\n",
                n,s[0],s[1],s[2],s[3],dev[0],dev[1]);
         fflush(flog);
      }
   }

   if (my_rank==0)
   {
      printf("\n");
      fclose(flog);
   }

   MPI_Finalize();
   exit(0);
}
