
/*******************************************************************************
*
* File time1.c
*
* Copyright (C) 2015 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Timing of the program ispt().
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "mpi.h"
#include "pmdb.h"
#include "utils.h"
#include "random.h"
#include "lattice.h"
#include "ispt.h"
#include "global.h"

static int my_rank,bc,Nf,order,iroot;
static int opv[(int)(PM_TYPES)];
static double *pv[(int)(PM_TYPES)];
static theory_t th;
static char *vtname[VTYPES+1]={"SOURCE","EXTERNAL",
                               "VS","VG","VGM","VFP","VQG",
                               "VSM","VGF","VQM",
                               "VTYPES"};
static char theory[NAME_SIZE];
static FILE *fin=NULL,*flog=NULL;


static void set_th(void)
{
   if (strcmp("PHI4",theory)==0)
      th=PHI4;
   else if (strcmp("YM",theory)==0)
      th=YM;
   else if (strcmp("QCD",theory)==0)
      th=QCD;
   else if (strcmp("LYM",theory)==0)
      th=LYM;
   else if (strcmp("LQCD",theory)==0)
      th=LQCD;
   else
      error(1,1,"set_th [time1.c]","Unknown theory");
}


static void read_theory_parms(void)
{
   find_section("Theory");
   read_line("theory","%s",theory);
   read_line("bc","%d",&bc);
   read_line("Nf","%d",&Nf);

   find_section("ISPT");
   read_line("order","%d",&order);
   read_line("iroot","%d",&iroot);
   set_th();
}


static void read_infile(void)
{
   int ith;

   if (my_rank==0)
   {
      fin=freopen("time1.in","r",stdin);
      error_root(fin==NULL,1,"read_infile [time1.c]",
                 "Unable to open input file");
      read_theory_parms();
      fclose(fin);
   }

   ith=(int)(th);
   MPI_Bcast(&ith,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&bc,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&Nf,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&order,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&iroot,1,MPI_INT,0,MPI_COMM_WORLD);
   th=(theory_t)(ith);
}


static void default_pv(void)
{
   int i,pmax;
   double *p;

   pmax=(int)(PM_TYPES);
   p=malloc(pmax*(order+1)*sizeof(*p));

   for (i=0;i<(pmax*(order+1));i++)
      p[i]=0.0;

   for (i=0;i<pmax;i++)
   {
      opv[i]=order;
      pv[i]=p;
      p+=order+1;
   }

   pv[(int)(C0)][0]=1.0;
   pv[(int)(LAMBDA)][0]=1.0;
   pv[(int)(MASS)][0]=0.0;
   pv[(int)(CG)][0]=1.0;
   pv[(int)(CGP)][0]=1.0;
   pv[(int)(CSW)][0]=1.0;
   pv[(int)(CF)][0]=1.0;
   pv[(int)(CFP)][0]=1.0;
}


static void print_tp(tupel_t *tp)
{
   int n,*i,*im;

   n=(*tp).n;

   if (n>0)
   {
      i=(*tp).i;
      im=i+n;

      printf("[%d",i[0]);
      i+=1;

      for (;i<im;i++)
         printf(",%d",i[0]);

      printf("]");
   }
}


int main(int argc,char *argv[])
{
   int n,i;
   int icl,ncl,nall,nw[4];
   double wt1,wt2,wdt,wtav,wtall;
   sl3_alg_dble **wad;
   rcls_t *rc;
   rlist_t *rl;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);
   set_error_file("time1.log",0);

   if (my_rank==0)
   {
      flog=freopen("time1.log","w",stdout);

      printf("\n");
      printf("Timing of the program valrt()\n");
      printf("-----------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      if (NPROC>1)
         printf("There are %d MPI processes\n",NPROC);
      else
         printf("There is 1 MPI process\n");

      if ((64*VOLUME*sizeof(double))<(4096*1024))
         printf("The local size of a complex gauge potential is %d KB\n",
                (int)((64*VOLUME*sizeof(double))/(1024)));
      else
         printf("The local size of a complex gauge potential is %d MB\n",
                (int)((64*VOLUME*sizeof(double))/(1024*1024)));

#if (defined x64)
#if (defined AVX)
      printf("Using AVX instructions\n");
#else
      printf("Using SSE3 instructions and 16 xmm registers\n");
#endif
#if (defined P3)
      printf("Assuming SSE prefetch instructions fetch 32 bytes\n");
#elif (defined PM)
      printf("Assuming SSE prefetch instructions fetch 64 bytes\n");
#elif (defined P4)
      printf("Assuming SSE prefetch instructions fetch 128 bytes\n");
#else
      printf("SSE prefetch instructions are not used\n");
#endif
#endif
      printf("\n");
      fflush(flog);
   }

   read_infile();
   set_th_parms(th,bc,Nf);
   default_pv();
   pv[(int)(LAMBDA)][0]=0.4567;
   set_ispt_parms(order,iroot,opv,pv);
   print_th_parms();
   print_ispt_parms();
   start_ranlux(0,12345);
   geometry();

   for (i=0;i<4;i++)
      nw[i]=0;

   if (th==PHI4)
      nw[0]=2;
   else
      nw[1]=2;

   init_ispt(nw);
   rc=rtree_classes(th);
   ncl=rc[0].cls[0].n;

   if (my_rank==0)
   {
      printf("There are %d classes of diagrams\n",ncl);

      printf("Counterterm vertices = ");

      for (i=0;i<rc[0].nvt;i++)
      {
         if (i>0)
            printf(", %s",vtname[(int)(rc[0].vt[i])]);
         else
            printf("%s",vtname[(int)(rc[0].vt[i])]);
      }

      printf("\n\n");
   }

   wad=reserve_wad(2);
   generate_rflds(wad[0]);
   nall=0;
   wtall=0.0;

   for (icl=0;icl<ncl;icl++)
   {
      MPI_Barrier(MPI_COMM_WORLD);
      wt1=MPI_Wtime();

      ispt(icl,wad[1]);

      wt2=MPI_Wtime();
      wdt=wt2-wt1;

      MPI_Reduce(&wdt,&wtav,1,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);
      wtav/=(double)(NPROC);

      rl=select_rtrees(th,icl);
      n=(*rl).n;

      if (my_rank==0)
      {
         printf("class=");
         print_tp(rc[0].cls[0].tpa[icl]);
         printf(", no of diagrams=%d\n",n);
         printf("Total time: %.1e sec\n",wtav);
         printf("Time per diagram: %.1e sec\n\n",wtav/(double)(n));
         fflush(flog);
      }

      nall+=n;
      wtall+=wtav;
   }

   if (my_rank==0)
   {
      printf("\n");
      printf("Total number of diagrams=%d\n",nall);
      printf("Total time: %.1e sec\n",wtall);
      printf("Time per diagram: %.1e sec\n\n",wtall/(double)(nall));
      fflush(flog);
   }

   release_wad();
   stop_ispt();

   if (my_rank==0)
      fclose(flog);

   MPI_Finalize();
   exit(0);
}
