
/*******************************************************************************
*
* File check2.c
*
* Copyright (C) 2014 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Reordering and comparison of branch diagrams.
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "utils.h"
#include "trees.h"

static int nvs=0,nls;
static int *rk1s,*rk2s;
static branch_t *brs;
static diagram_t *dm1s,*dm2s;


static theory_t theory(char *name)
{
   if (strcmp("PHI4",name)==0)
      return PHI4;
   else if (strcmp("YM",name)==0)
      return YM;
   else if (strcmp("QCD",name)==0)
      return QCD;
   else if (strcmp("LYM",name)==0)
      return LYM;
   else if (strcmp("LQCD",name)==0)
      return LQCD;
   else
      error(1,1,"theory [check2.c]","Unknown theory");

   return THEORIES;
}


static void alloc_aux(theory_t th,int nbr)
{
   int nv,ibr;

   if (nvs>0)
   {
      free(rk1s);
      free_branch(brs);
      free_dgm(dm1s);
      free_dgm(dm2s);
   }

   nvs=0;

   for (ibr=0;ibr<nbr;ibr++)
   {
      nv=get_nv(th,ibr);

      if (nv>nvs)
         nvs=nv;
   }

   nls=nvs-1;

   rk1s=malloc(2*nvs*sizeof(*rk1s));
   rk2s=rk1s+nvs;

   error(rk1s==NULL,1,"alloc_aux [check2.c]",
         "Unable to allocate auxiliary arrays");

   brs=alloc_branch(nvs);
   dm1s=alloc_dgm(nvs,nls);
   dm2s=alloc_dgm(nvs,nls);
}


static void print_dgms(void)
{
   int nv,nl,i;
   int (*v1)[2],(*v2)[2];
   line_t *l1,*l2;

   nv=(*dm1s).nv;
   nl=(*dm1s).nl;
   v1=(*dm1s).v;
   l1=(*dm1s).l;
   v2=(*dm2s).v;
   l2=(*dm2s).l;

   error((nv!=(*dm2s).nv)||(nl!=(*dm2s).nl),1,"print_dgms [check2.c]",
         "Unexpected diagram data");

   for (i=0;i<nv;i++)
   {
      printf(" %2d  %2d  %d,%d",i,rk1s[i],v1[i][0],v1[i][1]);
      printf(" | %2d  %d,%d",rk2s[i],v2[i][0],v2[i][1]);
      printf("\n");
   }

   printf("\n");

   for (i=0;i<nl;i++)
   {
      printf("%2d=>%2d ",l1[i].iv[0],l1[i].iv[1]);

      if (l1[i].type==SCALAR)
         printf("SCALAR   ");
      else if (l1[i].type==GLUON)
         printf("GLUON    ");
      else if (l1[i].type==GHOST)
         printf("GHOST    ");
      else if (l1[i].type==ANTIGHOST)
         printf("ANTIGHOST");
      else if (l1[i].type==QUARK)
         printf("QUARK    ");
      else if (l1[i].type==ANTIQUARK)
         printf("ANTIQUARK");
      else
         error(1,1,"print_dgms [check2.c]","Unknown line type");

      printf(" | %2d=>%2d ",l2[i].iv[0],l2[i].iv[1]);

      if (l2[i].type==SCALAR)
         printf("SCALAR   ");
      else if (l2[i].type==GLUON)
         printf("GLUON    ");
      else if (l2[i].type==GHOST)
         printf("GHOST    ");
      else if (l2[i].type==ANTIGHOST)
         printf("ANTIGHOST");
      else if (l2[i].type==QUARK)
         printf("QUARK    ");
      else if (l2[i].type==ANTIQUARK)
         printf("ANTIQUARK");
      else
         error(1,1,"print_dgms [check2.c]","Unknown line type");

      printf("\n");
   }

   printf("\n");
}


static void check_inverse_branches(theory_t th,int nbr)
{
   int ibr,ibs,k;

   for (ibr=0;ibr<nbr;ibr++)
   {
      (*brs).nv=nvs;
      (*dm1s).nv=nvs;
      (*dm1s).nl=nls;
      (*dm2s).nv=nvs;
      (*dm2s).nl=nls;

      get_branch(th,ibr,brs);
      branch2dgm(brs,dm1s);
      reorder_dgm(dm1s,rk1s);

      ibs=inverse_branch(th,ibr);
      get_branch(th,ibs,brs);
      branch2dgm(brs,dm2s);
      reorder_dgm(dm2s,rk2s);

      k=cmp_dgm(dm1s,rk1s,dm2s,rk2s);

      if (k!=1)
      {
         print_dgms();
         error(1,1,"check_inverse_branches [check2.c]",
               "Branch and its inverse are not the same diagram");
      }
   }
}


static void check_unequal_branches(theory_t th,int nbr)
{
   int ib1,ib2,ibs,k;

   for (ib1=0;ib1<nbr;ib1++)
   {
      (*brs).nv=nvs;
      (*dm1s).nv=nvs;
      (*dm1s).nl=nls;

      get_branch(th,ib1,brs);
      branch2dgm(brs,dm1s);
      reorder_dgm(dm1s,rk1s);

      ibs=inverse_branch(th,ib1);

      for (ib2=(ib1+1);ib2<nbr;ib2++)
      {
         if (ib2!=ibs)
         {
            (*brs).nv=nvs;
            (*dm2s).nv=nvs;
            (*dm2s).nl=nls;

            get_branch(th,ib2,brs);
            branch2dgm(brs,dm2s);
            reorder_dgm(dm2s,rk2s);

            k=cmp_dgm(dm1s,rk1s,dm2s,rk2s);

            if (k!=0)
            {
               print_dgms();
               error(1,1,"check_unequal_branches [check2.c]",
                     "Some unequal branches give the same diagram");
            }
         }
      }
   }
}


int main(void)
{
   int order,nbr;
   char name[5];
   theory_t th;

   printf("\n");
   printf("Reordering and comparison of branch diagrams\n");
   printf("--------------------------------------------\n\n");

   while (1)
   {
      printf("Specify theory and maximal order: ");
      (void)scanf("%4s %d",name,&order);

      th=theory(name);
      nbr=branches(th,order);
      printf("\n");
      printf("Total number of branches = %d\n\n",nbr);

      alloc_aux(th,nbr);
      check_inverse_branches(th,nbr);
      check_unequal_branches(th,nbr);
   }

   exit(0);
}
