
/*******************************************************************************
*
* File check3.c
*
* Copyright (C) 2015 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Check of the gauge damping term.
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "sl3.h"
#include "random.h"
#include "pmdb.h"
#include "utils.h"
#include "lattice.h"
#include "su3fcts.h"
#include "aflds.h"
#include "uflds.h"
#include "linalg.h"
#include "wflow.h"
#include "global.h"

#define N0 (NPROC0*L0)
#define N1 (NPROC1*L1)
#define N2 (NPROC2*L2)
#define N3 (NPROC3*L3)


static void new_fld(sl3_alg_dble *ad)
{
   if (cpr[0]==(NPROC0-1))
      gauss_dble((double*)(ad),16*(4*VOLUME+7*(BNDRY/4)+3));
   else
      gauss_dble((double*)(ad),16*(4*VOLUME+7*(BNDRY/4)));

   mulr_calg(4*VOLUME,0.1,ad,ad);
   set_adbc(ad);
}


static int is_zero(su3_dble *ud)
{
   int it;

   it=1;
   it&=((*ud).c11.re==0.0);
   it&=((*ud).c11.im==0.0);
   it&=((*ud).c12.re==0.0);
   it&=((*ud).c12.im==0.0);
   it&=((*ud).c13.re==0.0);
   it&=((*ud).c13.im==0.0);

   it&=((*ud).c21.re==0.0);
   it&=((*ud).c21.im==0.0);
   it&=((*ud).c22.re==0.0);
   it&=((*ud).c22.im==0.0);
   it&=((*ud).c23.re==0.0);
   it&=((*ud).c23.im==0.0);

   it&=((*ud).c31.re==0.0);
   it&=((*ud).c31.im==0.0);
   it&=((*ud).c32.re==0.0);
   it&=((*ud).c32.im==0.0);
   it&=((*ud).c33.re==0.0);
   it&=((*ud).c33.im==0.0);

   return it;
}


static int check_udbc(int n,su3_dble *ud)
{
   int it,is,i;
   int npts,*pts,*ptm;
   su3_dble *vd;

   it=1;

   if ((cpr[0]==0)&&(bc_type()==1))
   {
      pts=bnd_pts(&npts);
      pts+=(npts/2);
      ptm=pts+(npts/2);

      for (;pts<ptm;pts++)
      {
         vd=ud+(8*(pts[0]-(VOLUME/2))+2)*(n+1);

         for (i=0;i<(6*(n+1));i++)
         {
            it&=is_zero(vd);
            vd+=1;
         }
      }
   }

   if (cpr[0]==(NPROC0-1))
   {
      vd=ud+(4*VOLUME+7*(BNDRY/4))*(n+1);

      for (i=0;i<(3*(n+1));i++)
      {
         it&=is_zero(vd);
         vd+=1;
      }
   }

   if (NPROC>1)
   {
      is=it;
      MPI_Allreduce(&is,&it,1,MPI_INT,MPI_MIN,MPI_COMM_WORLD);
   }

   return it;
}


int main(int argc,char *argv[])
{
   int my_rank,order,bc;
   int ns,rule,m,i;
   double eps,dev;
   complex_dble *sw0,*sw1,*sw2,z;
   complex_dble *plaq_sp0,*plaq_sp1,*plaq_sp2;
   complex_dble *plaq_tm0,*plaq_tm1,*plaq_tm2;
   sl3_alg_dble **wad;
   su3_dble **wud,*ud1,*ud2;
   FILE *flog=NULL,*fin=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("check3.log","w",stdout);
      fin=freopen("check3.in","r",stdin);

      printf("\n");
      printf("Check of the gauge damping term\n");
      printf("-------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      find_section("Theory");
      read_line("order","%d",&order);
      read_line("bc","%d\n",&bc);

      find_section("Wilson flow");
      read_line("ns","%d\n",&ns);
      read_line("eps","%lf\n",&eps);
      read_line("rule","%d\n",&rule);
      fclose(fin);

      error_root((rule<0)||(rule>3),1,"main [check3.c]",
                 "rule must be 1,2 or 3");
   }

   MPI_Bcast(&order,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&bc,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&ns,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&eps,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&rule,1,MPI_INT,0,MPI_COMM_WORLD);

   set_th_parms(LYM,bc,0);
   print_th_parms();

   if (my_rank==0)
   {
      printf("order = %d\n",order);
      printf("ns = %d\n",ns);
      printf("eps = %.3e\n",eps);

      if (rule==1)
         printf("Using the Euler integrator\n\n");
      else if (rule==2)
         printf("Using the 2nd order RK integrator\n\n");
      else
         printf("Using the 3rd order RK integrator\n\n");

      printf("Random initial fields\n\n");
      fflush(flog);
   }

   start_ranlux(0,12345);
   geometry();
   m=order+1;
   alloc_wad(3*m);
   alloc_wud(2*m);
   alloc_wcd(m);
   sw0=malloc(9*m*sizeof(*sw0));
   sw1=sw0+m;
   sw2=sw0+2*m;
   plaq_sp0=sw0+3*m;
   plaq_sp1=sw0+4*m;
   plaq_sp2=sw0+5*m;
   plaq_tm0=sw0+6*m;
   plaq_tm1=sw0+7*m;
   plaq_tm2=sw0+8*m;
   error(sw0==NULL,1,"main [check3.c]","Unable to allocate auxiliary array");

   wad=reserve_wad(m);

   for (i=0;i<m;i++)
      new_fld(wad[i]);

   wud=reserve_wud(2*m);
   ud1=wud[0];
   ud2=wud[m];
   a2uflds(order,wad,ud1);
   release_wad();

   error(check_udbc(order,ud1)!=1,1,"main [check3.c]",
         "Boundary conditions are not correctly set");
   Splaq(order,ud1,sw0);
   Eplaq(order,ud1,plaq_sp0,plaq_tm0);
   cm3x3_assign(4*VOLUME*m,ud1,ud2);

   if (rule==1)
      fwd_euler(ns,eps,0,order,ud2);
   else if (rule==2)
      fwd_rk2(ns,eps,0,order,ud2);
   else
      fwd_rk3(ns,eps,0,order,ud2);

   error(check_udbc(order,ud2)!=1,1,"main [check3.c]",
         "Boundary conditions are not preserved (idmp=0)");
   Splaq(order,ud2,sw1);
   Eplaq(order,ud2,plaq_sp1,plaq_tm1);

   cm3x3_assign(4*VOLUME*m,ud1,ud2);

   if (rule==1)
      fwd_euler(ns,eps,1,order,ud2);
   else if (rule==2)
      fwd_rk2(ns,eps,1,order,ud2);
   else
      fwd_rk3(ns,eps,1,order,ud2);

   error(check_udbc(order,ud2)!=1,1,"main [check3.c]",
         "Boundary conditions are not preserved (idmp=1)");
   Splaq(order,ud2,sw2);
   Eplaq(order,ud2,plaq_sp2,plaq_tm2);

   if (my_rank==0)
   {
      printf("Before integration:\n");

      for (i=0;i<m;i++)
         printf("i = %d, Splaq = % .8e,% .8e\n",i,sw0[i].re,sw0[i].im);

      printf("\n");
      printf("After integration:\n");

      for (i=0;i<m;i++)
         printf("i = %d, Splaq[idmp=0] = % .8e,% .8e\n",
                i,sw1[i].re,sw1[i].im);

      printf("\n");

      for (i=0;i<m;i++)
      {
         z.re=sw2[i].re-sw1[i].re;
         z.im=sw2[i].im-sw1[i].im;
         dev=(z.re*z.re+z.im*z.im)/(sw2[i].re*sw2[i].re+sw2[i].im*sw2[i].im);

         printf("i = %d, Splaq[idmp=1] = % .8e,% .8e, "
                "relative deviation = %.1e\n",
                i,sw2[i].re,sw2[i].im,sqrt(dev));
      }

      printf("\n");

      for (i=0;i<m;i++)
         printf("i = %d, plaq_sp[idmp=0] = % .8e,% .8e\n",
                i,plaq_sp1[i].re,plaq_sp1[i].im);

      printf("\n");

      for (i=0;i<m;i++)
      {
         z.re=plaq_sp2[i].re-plaq_sp1[i].re;
         z.im=plaq_sp2[i].im-plaq_sp1[i].im;
         dev=(z.re*z.re+z.im*z.im)/
            (plaq_sp2[i].re*plaq_sp2[i].re+plaq_sp2[i].im*plaq_sp2[i].im);

         printf("i = %d, plaq_sp[idmp=1] = % .8e,% .8e, "
                "relative deviation = %.1e\n",
                i,plaq_sp2[i].re,plaq_sp2[i].im,sqrt(dev));
      }

      printf("\n");

      for (i=0;i<m;i++)
         printf("i = %d, plaq_tm[idmp=0] = % .8e,% .8e\n",
                i,plaq_tm1[i].re,plaq_tm1[i].im);

      printf("\n");

      for (i=0;i<m;i++)
      {
         z.re=plaq_tm2[i].re-plaq_tm1[i].re;
         z.im=plaq_tm2[i].im-plaq_tm1[i].im;
         dev=(z.re*z.re+z.im*z.im)/
            (plaq_tm2[i].re*plaq_tm2[i].re+plaq_tm2[i].im*plaq_tm2[i].im);

         printf("i = %d, plaq_tm[idmp=1] = % .8e,% .8e, "
                "relative deviation = %.1e\n",
                i,plaq_tm2[i].re,plaq_tm2[i].im,sqrt(dev));
      }

      printf("\n");

      fclose(flog);
   }

   MPI_Finalize();
   exit(0);
}
