
/*******************************************************************************
*
* File aprop.c
*
* Copyright (C) 2015, 2017 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Gauge-field propagator and generation of random gauge potentials.
*
* The externally accessible functions are
*
*   void aprop(sl3_alg_dble *src,sl3_alg_dble *fld)
*     Applies the gauge-field propagator to the source src and assigns
*     the calculated field to fld.
*
*   void random_asrc(sl3_alg_dble *fld)
*     Generates a new Gaussian random gauge potential representing the free
*     theory and assigns the field to fld.
*
* Gauge potentials are locally stored as described in main/README.global. In
* particular, the input and output fields must be allocated with size equal to
* at least 4*VOLUME+7*(BNDRY/4) (+3 if cpr[0]=NPROC0-1). The program aprop()
* leaves the input field unchanged unless it overlaps the output field (which
* is permissible).
*
* The programs in this module act globally and must be called on all MPI
* processes simultaneously.
*
*******************************************************************************/

#define APROP_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "sl3.h"
#include "su3fcts.h"
#include "utils.h"
#include "pmdb.h"
#include "lattice.h"
#include "linalg.h"
#include "dft.h"
#include "aflds.h"
#include "global.h"

static int idp[4],init=0;
static double *phat[4];
static const sl3_alg_dble ad0={{0.0,0.0},{0.0,0.0},{0.0,0.0},{0.0,0.0},
                               {0.0,0.0},{0.0,0.0},{0.0,0.0},{0.0,0.0}};
static sl3_alg_dble pA ALIGNED16;
static sl3_alg_dble *sad[4];


static void alloc_sad(void)
{
   int vol,i;

   if (cpr[0]==(NPROC0-1))
      vol=VOLUME+L1*L2*L3;
   else
      vol=VOLUME;

   sad[0]=amalloc(4*vol*sizeof(*sad[0]),ALIGN);
   sad[1]=sad[0]+vol;
   sad[2]=sad[1]+vol;
   sad[3]=sad[2]+vol;

   error(sad[0]==NULL,1,"alloc_sad [aprop.c]",
         "Unable to allocate field array");

   for (i=0;i<(4*vol);i++)
      sad[0][i]=ad0;
}


static void set_idp(void)
{
   int bc,mu;
   int n[4],id[4],nx[4];
   th_parms_t th;
   ispt_parms_t isp;

   th=th_parms();
   isp=ispt_parms();

   error(th.th==THEORIES,1,"set_idp [aprop.c]",
         "Theory parameters are not set");
   error((isp.order==0),1,"set_idp [aprop.c]",
         "ISPT parameters are not set");
   error(iup[0][0]==0,1,"set_idp [aprop.c]",
         "Geometry arrays are not set");

   n[0]=NPROC0*L0;
   n[1]=NPROC1*L1;
   n[2]=NPROC2*L2;
   n[3]=NPROC3*L3;

   if (cpr[0]==(NPROC0-1))
      nx[0]=L0+1;
   else
      nx[0]=L0;

   nx[1]=L1;
   nx[2]=L2;
   nx[3]=L3;
   bc=bc_type();

   if (bc==1)
   {
      for (mu=0;mu<4;mu++)
      {
         if (mu==0)
         {
            id[0]=set_dft_parms(COS,n[0],1,0);
            id[1]=set_dft_parms(EXP,n[1],0,0);
            id[2]=set_dft_parms(EXP,n[2],0,0);
            id[3]=set_dft_parms(EXP,n[3],0,0);
         }
         else
         {
            id[0]=set_dft_parms(SIN,n[0],0,0);
            id[1]=set_dft_parms(EXP,n[1],(mu==1),0);
            id[2]=set_dft_parms(EXP,n[2],(mu==2),0);
            id[3]=set_dft_parms(EXP,n[3],(mu==3),0);
         }

         idp[mu]=set_dft4d_parms(id,nx,8);
      }
   }
   else
   {
      for (mu=0;mu<4;mu++)
      {
         if (mu==0)
         {
            id[0]=set_dft_parms(SIN,n[0],1,1);
            id[1]=set_dft_parms(EXP,n[1],0,0);
            id[2]=set_dft_parms(EXP,n[2],0,0);
            id[3]=set_dft_parms(EXP,n[3],0,0);
         }
         else
         {
            id[0]=set_dft_parms(COS,n[0],0,1);
            id[1]=set_dft_parms(EXP,n[1],(mu==1),0);
            id[2]=set_dft_parms(EXP,n[2],(mu==2),0);
            id[3]=set_dft_parms(EXP,n[3],(mu==3),0);
         }

         idp[mu]=set_dft4d_parms(id,nx,8);
      }
   }
}


static void set_phat(void)
{
   int l[4],mu,k,k0,bc;
   double r[4],pi;

   phat[0]=malloc((L0+L1+L2+L3)*sizeof(double));
   error(phat[0]==NULL,1,"set_phat [aprop.c]",
         "Unable to allocate auxiliarly array");
   phat[1]=phat[0]+L0;
   phat[2]=phat[1]+L1;
   phat[3]=phat[2]+L2;

   pi=4.0*atan(1.0);
   r[0]=pi/(double)(2*NPROC0*L0);
   r[1]=pi/(double)(NPROC1*L1);
   r[2]=pi/(double)(NPROC2*L2);
   r[3]=pi/(double)(NPROC3*L3);

   l[0]=L0;
   l[1]=L1;
   l[2]=L2;
   l[3]=L3;

   bc=bc_type();

   for (mu=0;mu<4;mu++)
   {
      k0=cpr[mu]*l[mu];

      if ((bc==1)||(mu>0))
      {
         for (k=0;k<l[mu];k++)
            phat[mu][k]=2.0*sin((double)(k+k0)*r[mu]);
      }
      else
      {
        for (k=0;k<l[mu];k++)
           phat[mu][k]=2.0*sin(((double)(k+k0)+0.5)*r[mu]);
      }
   }
}


static void correct_bnd(double r)
{
   int ix,mu;
   sl3_alg_dble *ad;

   for (mu=1;mu<4;mu++)
   {
      ad=sad[mu];

      for (ix=0;ix<(L1*L2*L3);ix++)
      {
         sl3_mulr(&r,ad,ad);
         ad+=1;
      }
   }
}


static void apply_prop0(void)
{
   int bc,ik,n,mu;
   double ph[4],phsq,r0,r1,r2;
   ispt_parms_t isp;

   bc=bc_type();
   isp=ispt_parms();
   r0=1.0/isp.pv[(int)(LAMBDA)][0];
   r1=r0-1.0;

   for (ik=0;ik<VOLUME;ik++)
   {
      n=ik;
      ph[3]=phat[3][n%L3];
      n/=L3;
      ph[2]=phat[2][n%L2];
      n/=L2;
      ph[1]=phat[1][n%L1];
      ph[0]=phat[0][n/L1];

      phsq=1.0/(ph[0]*ph[0]+ph[1]*ph[1]+ph[2]*ph[2]+ph[3]*ph[3]);

      if ((ik>=(L1*L2*L3))||(bc!=1)||(cpr[0]>0))
      {
         sl3_mulr(ph,sad[0]+ik,&pA);
         sl3_mulr_add(ph+1,sad[1]+ik,&pA);
         sl3_mulr_add(ph+2,sad[2]+ik,&pA);
         sl3_mulr_add(ph+3,sad[3]+ik,&pA);

         for (mu=0;mu<4;mu++)
         {
            r2=ph[mu]*r1*phsq;
            sl3_mulr_add(&r2,&pA,sad[mu]+ik);
            sl3_mulr(&phsq,sad[mu]+ik,sad[mu]+ik);
         }
      }
      else
      {
         if ((ik>0)||(cpr[1]>0)||(cpr[2]>0)||(cpr[3]>0))
            r2=phsq;
         else
         {
            r2=(double)(NPROC0*L0);
            r2=r0*r2*r2;
         }

         sl3_mulr(&r2,sad[0]+ik,sad[0]+ik);
      }
   }
}


void aprop(sl3_alg_dble *src,sl3_alg_dble *fld)
{
   int mu;

   if (init==0)
   {
      alloc_sad();
      set_idp();
      set_phat();
      init=1;
   }

   gather_afld(src,sad);

   if ((bc_type()==2)&&(cpr[0]==0))
      correct_bnd(2.0);

   for (mu=0;mu<4;mu++)
      inv_dft4d(idp[mu],(complex_dble*)(sad[mu]),(complex_dble*)(sad[mu]));

   apply_prop0();

   for (mu=0;mu<4;mu++)
      dft4d(idp[mu],(complex_dble*)(sad[mu]),(complex_dble*)(sad[mu]));

   scatter_afld(sad,fld);

   if (cpr[0]==(NPROC0-1))
      set_calg2zero(3,fld+4*VOLUME+7*(BNDRY/4));
}


static void apply_prop1(void)
{
   int bc,ik,n,mu;
   double ph[4],phsq,r0,r1,r2;
   ispt_parms_t isp;

   bc=bc_type();
   isp=ispt_parms();
   r0=1.0/sqrt(isp.pv[(int)(LAMBDA)][0]);
   r1=r0-1.0;

   for (ik=0;ik<VOLUME;ik++)
   {
      n=ik;
      ph[3]=phat[3][n%L3];
      n/=L3;
      ph[2]=phat[2][n%L2];
      n/=L2;
      ph[1]=phat[1][n%L1];
      ph[0]=phat[0][n/L1];

      phsq=1.0/sqrt(ph[0]*ph[0]+ph[1]*ph[1]+ph[2]*ph[2]+ph[3]*ph[3]);

      if ((ik>=(L1*L2*L3))||(bc!=1)||(cpr[0]>0))
      {
         sl3_mulr(ph,sad[0]+ik,&pA);
         sl3_mulr_add(ph+1,sad[1]+ik,&pA);
         sl3_mulr_add(ph+2,sad[2]+ik,&pA);
         sl3_mulr_add(ph+3,sad[3]+ik,&pA);

         for (mu=0;mu<4;mu++)
         {
            r2=ph[mu]*r1*phsq*phsq;
            sl3_mulr_add(&r2,&pA,sad[mu]+ik);
            sl3_mulr(&phsq,sad[mu]+ik,sad[mu]+ik);
         }
      }
      else
      {
         if ((ik>0)||(cpr[1]>0)||(cpr[2]>0)||(cpr[3]>0))
            r2=phsq;
         else
            r2=r0*(double)(NPROC0*L0);

         sl3_mulr(&r2,sad[0]+ik,sad[0]+ik);
      }
   }
}


void random_asrc(sl3_alg_dble *fld)
{
   int mu;

   if (init==0)
   {
      alloc_sad();
      set_idp();
      set_phat();
      init=1;
   }

   random_calg(4*VOLUME,fld);
   set_adbc(fld);
   gather_afld(fld,sad);

   if ((bc_type()==2)&&(cpr[0]==0))
      correct_bnd(sqrt(2.0));

   for (mu=0;mu<4;mu++)
      inv_dft4d(idp[mu],(complex_dble*)(sad[mu]),(complex_dble*)(sad[mu]));

   apply_prop1();

   for (mu=0;mu<4;mu++)
      dft4d(idp[mu],(complex_dble*)(sad[mu]),(complex_dble*)(sad[mu]));

   scatter_afld(sad,fld);
}
