
/*******************************************************************************
*
* File cprop.c
*
* Copyright (C) 2015 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Ghost-field propagator and generation of random ghost fields.
*
*   void cprop(sl3_alg_dble *src,sl3_alg_dble *fld)
*     Applies the ghost-field propagator to the source src and assigns
*     the calculated field to fld.
*
*   void random_csrc(sl3_alg_dble *cbfld,sl3_alg_dble *cfld)
*     Generates a new pair of Gaussian random antighost and ghost fields
*     representing the free theory and assigns them to cbfld and cfld,
*     respectively.
*
* Ghost fields are locally stored as described in main/README.global and
* README.cflds. In particular, the input and output fields must be allocated
* with size equal to NSPIN or more. The program cprop() leaves the input
* field unchanged unless it overlaps the output field (which is permissible).
*
* Since the ghost-field propagator is symmetric in its arguments, the same
* program can be used for the propagation of ghost and antighost fields.
*
* The programs in this module act globally and must be called on all MPI
* processes simultaneously.
*
*******************************************************************************/

#define CPROP_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "sl3.h"
#include "su3fcts.h"
#include "utils.h"
#include "pmdb.h"
#include "lattice.h"
#include "linalg.h"
#include "dft.h"
#include "cflds.h"
#include "global.h"

#define N0 (NPROC0*L0)
#define N1 (NPROC1*L1)
#define N2 (NPROC2*L2)
#define N3 (NPROC3*L3)

static int idp,idsm,init=0;
static double *phat[4],rt[N0];
static const sl3_alg_dble cd0={{0.0,0.0},{0.0,0.0},{0.0,0.0},{0.0,0.0},
                               {0.0,0.0},{0.0,0.0},{0.0,0.0},{0.0,0.0}};
static sl3_alg_dble smJ,*scd;


static void alloc_scd(void)
{
   int vol,i;

   if (cpr[0]==(NPROC0-1))
      vol=VOLUME+L1*L2*L3;
   else
      vol=VOLUME;

   scd=amalloc(vol*sizeof(*scd),ALIGN);

   error(scd==NULL,1,"alloc_scd [cprop.c]",
         "Unable to allocate field array");

   for (i=0;i<vol;i++)
      scd[i]=cd0;
}


static void set_ids(void)
{
   idsm=init_hsum(16);
}


static void set_idp(void)
{
   int bc,n[4],id[4],nx[4];
   th_parms_t th;

   th=th_parms();

   error((th.th==THEORIES)||(iup[0][0]==0),1,"set_idp [cprop.c]",
         "Theory parameters or geometry arrays are not set");

   n[0]=N0;
   n[1]=N1;
   n[2]=N2;
   n[3]=N3;

   if (cpr[0]==(NPROC0-1))
      nx[0]=L0+1;
   else
      nx[0]=L0;

   nx[1]=L1;
   nx[2]=L2;
   nx[3]=L3;
   bc=bc_type();

   if (bc==1)
   {
      id[0]=set_dft_parms(SIN,n[0],0,0);
      id[1]=set_dft_parms(EXP,n[1],0,0);
      id[2]=set_dft_parms(EXP,n[2],0,0);
      id[3]=set_dft_parms(EXP,n[3],0,0);
   }
   else
   {
      id[0]=set_dft_parms(COS,n[0],0,1);
      id[1]=set_dft_parms(EXP,n[1],0,0);
      id[2]=set_dft_parms(EXP,n[2],0,0);
      id[3]=set_dft_parms(EXP,n[3],0,0);
   }

   idp=set_dft4d_parms(id,nx,8);
}


static void set_phat(void)
{
   int l[4],mu,k,k0,bc;
   double r[4],pi;

   phat[0]=malloc((L0+L1+L2+L3)*sizeof(double));
   error(phat[0]==NULL,1,"set_phat [cprop.c]",
         "Unable to allocate auxiliarly array");
   phat[1]=phat[0]+L0;
   phat[2]=phat[1]+L1;
   phat[3]=phat[2]+L2;

   pi=4.0*atan(1.0);
   r[0]=pi/(double)(2*N0);
   r[1]=pi/(double)(N1);
   r[2]=pi/(double)(N2);
   r[3]=pi/(double)(N3);

   l[0]=L0;
   l[1]=L1;
   l[2]=L2;
   l[3]=L3;

   bc=bc_type();

   for (mu=0;mu<4;mu++)
   {
      k0=cpr[mu]*l[mu];

      if ((bc==1)||(mu>0))
      {
         for (k=0;k<l[mu];k++)
            phat[mu][k]=2.0*sin((double)(k+k0)*r[mu]);
      }
      else
      {
         for (k=0;k<l[mu];k++)
            phat[mu][k]=2.0*sin(((double)(k+k0)+0.5)*r[mu]);
      }
   }
}


static void set_rt(void)
{
   int t;

   for (t=0;t<N0;t++)
      rt[t]=1.0-(double)(t)/(double)(N0);
}


static void gather_cfld(int bc,sl3_alg_dble *cd)
{
   int ix,iy,t;
   sl3_alg_dble X,ct0;

   if (bc==1)
   {
      avg_cd0(cd,&ct0);
      reset_hsum(idsm);

      for (ix=0;ix<VOLUME;ix++)
      {
         iy=ipt[ix];
         t=global_time(iy);

         if (t>0)
         {
            scd[ix]=cd[iy];
            _sl3_alg_mulr_assign(X,rt[t],cd[iy]);
            add_to_hsum(idsm,(double*)(&X));
         }
         else
         {
            scd[ix]=cd0;
            add_to_hsum(idsm,(double*)(&ct0));
         }
      }

      global_hsum(idsm,(double*)(&smJ));
   }
   else
   {
      for (ix=0;ix<VOLUME;ix++)
      {
         iy=ipt[ix];
         scd[ix]=cd[iy];
      }
   }
}


static void scatter_cfld(int bc,sl3_alg_dble *cd)
{
   int ix,iy,t;
   double r0,r1;

   if (bc==1)
   {
      r0=(double)(N0)/(double)(N1*N2*N3);

      for (ix=0;ix<VOLUME;ix++)
      {
         iy=ipt[ix];
         t=global_time(iy);

         cd[iy]=scd[ix];
         r1=rt[t]*r0;
         _sl3_alg_mulr_add_assign(cd[iy],r1,smJ);
      }
   }
   else
   {
      for (ix=0;ix<VOLUME;ix++)
      {
         iy=ipt[ix];
         cd[iy]=scd[ix];
      }
   }
}


static void correct_bnd(double r)
{
   int ix;
   sl3_alg_dble *cd;

   cd=scd;

   for (ix=0;ix<(L1*L2*L3);ix++)
   {
      sl3_mulr(&r,cd,cd);
      cd+=1;
   }
}


static void apply_prop0(void)
{
   int bc,ik,n;
   double ph[4],phsq;
   sl3_alg_dble *cd;

   bc=bc_type();
   cd=scd;

   for (ik=0;ik<VOLUME;ik++)
   {
      if ((ik>=(L1*L2*L3))||(cpr[0]>0)||(bc==2))
      {
         n=ik;
         ph[3]=phat[3][n%L3];
         n/=L3;
         ph[2]=phat[2][n%L2];
         n/=L2;
         ph[1]=phat[1][n%L1];
         ph[0]=phat[0][n/L1];

         phsq=1.0/(ph[0]*ph[0]+ph[1]*ph[1]+ph[2]*ph[2]+ph[3]*ph[3]);
         sl3_mulr(&phsq,cd,cd);
      }

      cd+=1;
   }
}


void cprop(sl3_alg_dble *src,sl3_alg_dble *fld)
{
   int bc;

   if (init==0)
   {
      alloc_scd();
      set_ids();
      set_idp();
      set_phat();
      set_rt();
      init=1;
   }

   bc=bc_type();
   gather_cfld(bc,src);

   if ((bc==2)&&(cpr[0]==0))
      correct_bnd(2.0);

   inv_dft4d(idp,(complex_dble*)(scd),(complex_dble*)(scd));
   apply_prop0();
   dft4d(idp,(complex_dble*)(scd),(complex_dble*)(scd));

   scatter_cfld(bc,fld);
}


void random_csrc(sl3_alg_dble *cbfld,sl3_alg_dble *cfld)
{
   int ix;
   sl3_alg_dble *cd;

   random_calg2(VOLUME,cbfld);
   set_cdbc(cbfld);
   cd=cfld;

   for (ix=0;ix<VOLUME;ix++)
   {
      (*cd).c1.re=-(*cbfld).c1.re;
      (*cd).c1.im=(*cbfld).c1.im;
      (*cd).c2.re=-(*cbfld).c2.re;
      (*cd).c2.im=(*cbfld).c2.im;

      (*cd).c3.re=-(*cbfld).c4.re;
      (*cd).c3.im=(*cbfld).c4.im;
      (*cd).c4.re=-(*cbfld).c3.re;
      (*cd).c4.im=(*cbfld).c3.im;

      (*cd).c5.re=-(*cbfld).c6.re;
      (*cd).c5.im=(*cbfld).c6.im;
      (*cd).c6.re=-(*cbfld).c5.re;
      (*cd).c6.im=(*cbfld).c5.im;

      (*cd).c7.re=-(*cbfld).c8.re;
      (*cd).c7.im=(*cbfld).c8.im;
      (*cd).c8.re=-(*cbfld).c7.re;
      (*cd).c8.im=(*cbfld).c7.im;

      cd+=1;
      cbfld+=1;
   }

   cprop(cfld,cfld);
}
