
/*******************************************************************************
*
* File ispt.c
*
* Copyright (C) 2015 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Generation of representative fields.
*
*   void init_ispt(int *nw)
*     Generates the trivializing map for the theory specified through the
*     parameter data base and brings the diagram lists in a form suitable
*     for numerical evaluation. The program then allocates and reserves
*     the workspace required for the computation of the diagrams through
*     the workspace facilities. A workspace for nw[0],..,nw[3] additional
*     scalar, gluon, ghost and quark fields is included in the allocation.
*
*   void stop_ispt(void)
*     Releases the workspaces previously reserved by init_ispt().
*
*   void ispt(int icl,void *fld)
*     Computes all diagrams in the class with index "icl" and assigns the
*     linear combination of their values to the field "fld". The random
*     source fields used are the ones last generated by generate_rflds()
*     [rflds.c].
*
* The programs in this module obtain the theory and ISPT parameters from the
* parameter data base. It is thus taken for granted that the parameters have
* been set through set_th_parms() and set_ispt_parms().
*
* Currently only the YM and LYM theories are supported. For the definition
* of the diagram classes see trees/rtree_gen.c.
*
* The workspace allocated by init_ispt() is sufficiently large to include the
* labeled random fields, the fields needed by valrt() and the temporary field
* used for the accumulation of the diagram values in the program ispt(). The
* nw[0],..,nw[3] additional fields are available to the calling programs.
*
* An error occurs if
*
*  - init_ispt() is called repeatedly without intervening calls of
*    stop_ispt(),
*
*  - a second call of init_ispt() requires a larger workspace than
*    the first call,
*
*  - ispt() is called without prior call of init_ispt(), or if
*    stop_ispt() was called since the last call of init_ispt(),
*
*  - the random source fields have not yet been generated when
*    ispt() is called.
*
* The programs in this module perform global operations and must be called
* simultaneously on all MPI processes.
*
*******************************************************************************/

#define ISPT_C

#include <stdlib.h>
#include <stdio.h>
#include "mpi.h"
#include "utils.h"
#include "pmdb.h"
#include "aflds.h"
#include "trees.h"
#include "contract.h"
#include "linalg.h"
#include "ispt.h"
#include "global.h"

static int init=0,order,iroot,nws[4];
static double nf;
static theory_t th;
static tplist_t *cls;


static void set_th(void)
{
   th_parms_t thp;
   ispt_parms_t isp;

   thp=th_parms();
   th=thp.th;
   nf=(double)(thp.Nf);
   isp=ispt_parms();
   order=isp.order;
   iroot=isp.iroot;

   error_root(th==THEORIES,1,"set_th [ispt.c]",
              "Theory parameters are not set");
   error_root(order==0,1,"set_th [ispt.c]",
              "ISPT parameters are not set");
   error_root((th!=YM)&&(th!=LYM),1,"set_th [ispt.c]",
              "Currently only the YM and LYM theories are supported");
}


static void ispt_wspace(int *nw)
{
   int ncl,icl;
   int n,k,l,nwmx[4];
   rtree_t **rta;
   rlist_t *rl;

   for (l=0;l<4;l++)
      nwmx[l]=0;

   ncl=(*cls).n;

   for (icl=0;icl<ncl;icl++)
   {
      rl=select_rtrees(th,icl);
      n=(*rl).n;
      rta=(*rl).rta;

      for (k=0;k<n;k++)
      {
         valrt_wspace(rta[k],nw);

         for (l=0;l<4;l++)
         {
            if (nw[l]>nwmx[l])
               nwmx[l]=nw[l];
         }
      }
   }

   for (l=0;l<4;l++)
      nw[l]=0;

   if (th==PHI4)
      nw[0]=1+order+nwmx[0];
   else
   {
      n=(order/2);
      nw[1]=1+n+nwmx[1];
      nw[2]=2*n+nwmx[2];

      if ((th==QCD)||(th==LQCD))
         nw[3]=2*n+nwmx[3];
   }
}


void init_ispt(int *nw)
{
   int i,ie,nwa[4];
   rcls_t *rc;

   error(init==2,1,"init_ispt [ispt.c]",
         "Second call of init_ispt() without intervening call of stop_ispt()");

   if (init==0)
   {
      set_th();
      (void)(tmap(th,order,iroot));
      (void)(rtrees(th,order));
      rc=rtree_classes(th);
      cls=(*rc).cls;
      ispt_wspace(nwa);

      for (i=0;i<4;i++)
      {
         nws[i]=nw[i];
         nwa[i]+=nw[i];
      }

      if (nwa[1]>0)
         alloc_wad(nwa[1]);
      if (nwa[2]>0)
         alloc_wcd(nwa[2]);
      if (nwa[3]>0)
         alloc_wsd(nwa[3]);

      init=1;
   }

   ie=0;

   for (i=0;i<4;i++)
      ie|=(nw[i]>nws[i]);

   error(ie!=0,1,"init_ispt [ispt.c]",
         "The previously allocated workspace is too small");

   reserve_rflds();
   init=2;
}


void stop_ispt(void)
{
   if (init==2)
   {
      release_rflds();
      init=1;
   }
}


void ispt(int icl,void *fld)
{
   int n,k,iprms[1];
   double c;
   void *vrt;
   sl3_alg_dble **wad;
   rtree_t **rta;
   cnum_t **cna;
   rlist_t *rl;

   error_root(init!=2,1,"ispt [ispt.c]",
              "init_ispt() has not yet been called");
   error_root((icl<0)||(icl>=(*cls).n),1,"ispt [ispt.c]",
              "Parameter 'icl' is out of range");

   if (NPROC>1)
   {
      iprms[0]=icl;
      MPI_Bcast(iprms,1,MPI_INT,0,MPI_COMM_WORLD);
      error(iprms[0]!=icl,1,"ispt [ispt.c]","Parameter 'icl' is not global");
   }

   set_calg2zero(4*VOLUME,(sl3_alg_dble*)(fld));

   if (cpr[0]==(NPROC0-1))
      set_calg2zero(3,(sl3_alg_dble*)(fld)+4*VOLUME+7*(BNDRY/4));

   rl=select_rtrees(th,icl);
   n=(*rl).n;
   rta=(*rl).rta;
   cna=(*rl).cna;

   wad=reserve_wad(1);
   vrt=(void*)(wad[0]);

   for (k=0;k<n;k++)
   {
      valrt(rta[k],vrt);
      c=eval_cnum(nf,cna[k]);
      mulr_add_calg(4*VOLUME,(sl3_alg_dble*)(fld),(sl3_alg_dble*)(vrt),c);
   }

   release_wad();

   aprop((sl3_alg_dble*)(fld),(sl3_alg_dble*)(fld));
}
