
/*******************************************************************************
*
* File rflds.c
*
* Copyright (C) 2015 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Generation and administration of the random source fields.
*
*   void reserve_rflds(void)
*     Reserves the memory space for the labeled random fields required
*     for ISPT through the workspace facility [utils/wspace.c]. No space
*     is reserved for the unlabeled random fields.
*
*   void release_rflds(void)
*     Releases the fields previously reserved by reserve_rflds().
*
*   void generate_rflds(void *fld0)
*     Generates a new set of random source fields, as required for ISPT,
*     and assigns the labeled random fields to the fields reserved by
*     reserve_rflds(). The unlabeled field is assigned to the field fld0.
*
*   rflds_t *rflds(void)
*     Returns the pointer to a statically allocated structure that contains
*     the addresses of the current random source fields and their numbers.
*     In the case of the unlabeled random field, the structure contains the
*     field address passed to generate_rflds() when this program was last
*     called.
*
* The programs in this module obtain the theory and ISPT parameters from the
* parameter data base. It is thus taken for granted that the parameters have
* been set through set_th_parms() and set_ispt_parms().
*
* The elements of an rflds_t structure are:
*
*   nflds       Array of the maximal numbers nflds[i] of field addresses
*               of type i=0,..,(int)(LTYPES)-1 included in the structure.
*               The labels of the random source fields of type i range
*               from 1 to nflds[i]-1.
*
*   flds        Array of the addresses flds[i][j] of the random fields of
*               type i with label j [j=1,..,nflds[i]-1, i as above]. The
*               addresses flds[i][0] are reserved for the unlabeled random
*               fields. Any unused addresses (flds[i][0] and flds[i] if
*               there are no random fields of type i) are set to NULL.
*
* The reserve and release programs do not allocate any fields and instead
* obtain the memory space for the random fields by calling the appropriate
* workspace functions.
*
* An error occurs if
*
*  - the random fields cannot be accommodated in the available workspace,
*
*  - reserve_rflds() is called more than once without intervening call
*    of release_rflds(),
*
*  - generate_rflds() is called when no space for the random fields is
*    reserved,
*
*  - rflds() is called when no space for the random fields is reserved
*    or generate_rflds() has not been called since reserve_rflds() was
*    last called.
*
* Except for rflds(), all programs in this module perform global operations
* and must be simultaneously called on all MPI processes.
*
*******************************************************************************/

#define RFLDS_C

#include <stdlib.h>
#include <stdio.h>
#include "mpi.h"
#include "utils.h"
#include "pmdb.h"
#include "aflds.h"
#include "cflds.h"
#include "ispt.h"

#define LTMAX (int)(LTYPES)

static int flag=0,order;
static theory_t th;
static rflds_t rfs;


static void alloc_flds(void)
{
   int n,i;
   void **p;

   n=0;

   for (i=0;i<LTMAX;i++)
      n+=rfs.nflds[i];

   p=malloc(n*sizeof(*p));
   error(p==NULL,1,"alloc_flds [rflds.c]",
         "Unable to allocate field pointer arrays");

   for (i=0;i<n;i++)
      p[i]=NULL;

   for (i=0;i<LTMAX;i++)
   {
      n=rfs.nflds[i];

      if (n>0)
      {
         rfs.flds[i]=p;
         p+=n;
      }
      else
         rfs.flds[i]=NULL;
   }

   flag=1;
}


void reserve_rflds(void)
{
   int n,i;
   void **flds,**wflds;
   th_parms_t thp;
   ispt_parms_t isp;

   thp=th_parms();
   th=thp.th;
   isp=ispt_parms();
   order=isp.order;

   error_root(th==THEORIES,1,"reserve_flds [rflds.c]",
              "Theory parameters are not set");
   error_root(order==0,1,"reserve_flds [rflds.c]",
              "ISPT parameters are not set");
   error_root((th!=YM)&&(th!=LYM),1,"reserve_flds [rflds.c]",
              "Currently only the pure gauge theory is supported");
   error_root(flag>1,1,"reserve_flds [rflds.c]",
              "Attempt to reserve fields a second time");

   if (flag==0)
   {
      for (i=0;i<LTMAX;i++)
         rfs.nflds[i]=0;

      if (th==PHI4)
         rfs.nflds[(int)(SCALAR)]=1+order;
      else
      {
         n=1+(order/2);

         rfs.nflds[(int)(GLUON)]=n;
         rfs.nflds[(int)(GHOST)]=n;
         rfs.nflds[(int)(ANTIGHOST)]=n;

         if ((th==QCD)||(th==LQCD))
         {
            rfs.nflds[(int)(QUARK)]=n;
            rfs.nflds[(int)(ANTIQUARK)]=n;
         }
      }

      alloc_flds();
   }

   if (th!=PHI4)
   {
      n=1+(order/2);

      if (n>1)
         wflds=(void**)(reserve_wad(n-1));

      flds=rfs.flds[(int)(GLUON)];
      flds[0]=NULL;

      for (i=1;i<n;i++)
         flds[i]=wflds[i-1];

      if (n>1)
         wflds=(void**)(reserve_wcd(2*n-2));

      flds=rfs.flds[(int)(GHOST)];
      flds[0]=NULL;

      for (i=1;i<n;i++)
         flds[i]=wflds[i-1];

      flds=rfs.flds[(int)(ANTIGHOST)];
      flds[0]=NULL;

      for (i=1;i<n;i++)
         flds[i]=wflds[n+i-2];

      if ((th==QCD)||(th==LQCD))
      {
         if (n>1)
            wflds=(void**)(reserve_wsd(2*n-2));

         flds=rfs.flds[(int)(QUARK)];
         flds[0]=NULL;

         for (i=1;i<n;i++)
            flds[i]=wflds[i-1];

         flds=rfs.flds[(int)(ANTIQUARK)];
         flds[0]=NULL;

         for (i=1;i<n;i++)
            flds[i]=wflds[n+i-2];
      }
   }

   flag=2;
}


void release_rflds()
{
   int n,i,j,*nflds;

   if (flag>1)
   {
      nflds=rfs.nflds;

      if (nflds[(int)(GLUON)]>1)
         release_wad();

      if (nflds[(int)(GHOST)]>1)
         release_wcd();

      if (nflds[(int)(QUARK)]>1)
         release_wsd();

      for (i=0;i<LTMAX;i++)
      {
         n=nflds[i];

         for (j=0;j<n;j++)
            rfs.flds[i][j]=NULL;
      }

      flag=1;
   }
}


void generate_rflds(void *fld0)
{
   int i,n,*nflds;
   void ***flds,**aflds,**cflds,**cbflds;

   error_root(flag<2,1,"generate_rflds [rflds.c]",
              "Currently no space is reserved for the random fields");

   nflds=rfs.nflds;
   flds=rfs.flds;

   if (th!=PHI4)
   {
      n=nflds[(int)(GLUON)];
      aflds=flds[(int)(GLUON)];
      aflds[0]=fld0;

      for (i=0;i<n;i++)
         random_asrc((sl3_alg_dble*)(aflds[i]));

      n=nflds[(int)(GHOST)];
      cflds=flds[(int)(GHOST)];
      cbflds=flds[(int)(ANTIGHOST)];

      for (i=1;i<n;i++)
         random_csrc((sl3_alg_dble*)(cbflds[i]),(sl3_alg_dble*)(cflds[i]));
   }

   flag=3;
}


rflds_t *rflds(void)
{
   error_loc(flag!=3,1,"rflds [rflds.c]",
             "The random fields first need to be generated");

   return &rfs;
}
