
/*******************************************************************************
*
* File sl3alg.c
*
* Copyright (C) 2015 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Functions for fields with values in the complexified Lie algebra of SU(3).
*
*   void random_calg(int vol,sl3_alg_dble *X)
*     Initializes the elements of the array X to random values in su(3)
*     with distribution proportional to exp{-(X,X)/2}.
*
*   void random_calg2(int vol,sl3_alg_dble *X)
*     Initializes the elements of the array X to random values in sl(3,C)
*     with distribution proportional to exp{-(X,X)}.
*
*   double norm_square_calg(int vol,int icom,sl3_alg_dble *X)
*     Returns the square of the norm of the field X.
*
*   complex_dble scalar_prod_calg(int vol,int icom,sl3_alg_dble *X,
*                                 sl3_alg_dble *Y)
*     Returns the scalar product of the fields X and Y.
*
*   void set_calg2zero(int vol,sl3_alg_dble *X)
*     Sets the elements of the array X to zero.
*
*   void assign_calg2calg(int vol,sl3_alg_dble *X,sl3_alg_dble *Y)
*     Assigns the field X to the field Y.
*
*   void assign_calgdag2calg(int vol,sl3_alg_dble *X,sl3_alg_dble *Y)
*     Assigns the Hermitian conjugate of the field X to the field Y.
*     If Y=X (which is permissible), X is replaced by its Hermitian
*     conjugate.
*
*   void add_calg(int vol,sl3_alg_dble *X,sl3_alg_dble *Y)
*     Adds the field X to Y.
*
*   void sub_calg(int vol,sl3_alg_dble *X,sl3_alg_dble *Y)
*     Subtracts the field X from Y.
*
*   void mulr_calg(int vol,double r,sl3_alg_dble *X,sl3_alg_dble *Y)
*     Assigns r*X to Y.
*
*   void mulr_add_calg(int vol,sl3_alg_dble *X,sl3_alg_dble *Y,double r)
*     Replaces the field X by X+r*Y.
*
*   void mulc_add_calg(int vol,sl3_alg_dble *X,sl3_alg_dble *Y,complex_dble z)
*     Replaces the field X by X+z*Y.
*
* The elements X of sl(3,C) are traceless complex 3x3 matrices. They are
* represented by sl3_alg_dble structures with complex elements c1,..,c8
* through
*
*  X_11=c1, X_22=c2, X_33=-c1-c2,
*
*  X_12=c3, X_21=c4, X_13=c5, X_31=c6, X_23=c7, X_32=c8.
*
* Matrices in the Lie algebra su(3) of SU(3) are instead represented by
* su3_alg_dble structures (see liealg.c). A matrix Y in su(3) is also
* a matrix X in sl(3,C). The associated data structures are related by
*
*  X.c1=i*(Y.c1+Y.c2),
*  X.c2=i*(Y.c2-2*Y.c1),
*  X.c3= Y.c3+i*Y.c4,
*  X.c4=-Y.c3+i*Y.c4,
*  X.c5= Y.c5+i*Y.c6,
*  X.c6=-Y.c5+i*Y.c6,
*  X.c7= Y.c7+i*Y.c8,
*  X.c8=-Y.c7+i*Y.c8.
*
* The scalar product of any two elements X,Y of sl(3,C) is defined by
*
*  (X,Y)=2*tr{X^dag*Y}
*
* and the square norm of X is
*
*  (X,X)=2*tr{X^dag*X},
*
* X^dag being the Hermitian conjugate of X.
*
* All programs in this module operate on arrays of Lie algebra elements whose
* base address is passed through the arguments. The length of the arrays is
* specified by the parameter vol.
*
* Square norms and scalar products are globally summed if the parameter icom
* is set to 1. The calculated values are then guaranteed to be exactly the
* same on all MPI processes. Otherwise the local sum is returned.
*
*******************************************************************************/

#define SL3ALG_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "sl3.h"
#include "utils.h"
#include "su3fcts.h"
#include "linalg.h"
#include "global.h"

static int idsmr,idsmz,init=0;
static su3_alg_dble Xr[2] ALIGNED16;


static void set_ids(void)
{
   idsmr=init_hsum(1);
   idsmz=init_hsum(2);
   init=1;
}


void random_calg(int vol,sl3_alg_dble *X)
{
   sl3_alg_dble *Xm;

   Xm=X+vol;

   for (;X<Xm;X++)
   {
      random_alg(1,Xr);

      (*X).c1.re=0.0;
      (*X).c1.im=Xr[0].c1+Xr[0].c2;
      (*X).c2.re=0.0;
      (*X).c2.im=Xr[0].c2-2.0*Xr[0].c1;

      (*X).c3.re=Xr[0].c3;
      (*X).c3.im=Xr[0].c4;
      (*X).c4.re=-Xr[0].c3;
      (*X).c4.im=Xr[0].c4;

      (*X).c5.re=Xr[0].c5;;
      (*X).c5.im=Xr[0].c6;
      (*X).c6.re=-Xr[0].c5;
      (*X).c6.im=Xr[0].c6;

      (*X).c7.re=Xr[0].c7;
      (*X).c7.im=Xr[0].c8;
      (*X).c8.re=-Xr[0].c7;
      (*X).c8.im=Xr[0].c8;
   }
}


void random_calg2(int vol,sl3_alg_dble *X)
{
   double r;
   sl3_alg_dble *Xm;

   r=1.0/sqrt(2.0);
   Xm=X+vol;

   for (;X<Xm;X++)
   {
      random_alg(2,Xr);

      (*X).c1.re=r*(Xr[1].c1+Xr[1].c2);
      (*X).c1.im=r*(Xr[0].c1+Xr[0].c2);
      (*X).c2.re=r*(Xr[1].c2-2.0*Xr[1].c1);
      (*X).c2.im=r*(Xr[0].c2-2.0*Xr[0].c1);

      (*X).c3.re=r*(Xr[0].c3+Xr[1].c4);
      (*X).c3.im=r*(Xr[0].c4-Xr[1].c3);
      (*X).c4.re=r*(Xr[1].c4-Xr[0].c3);
      (*X).c4.im=r*(Xr[1].c3+Xr[0].c4);

      (*X).c5.re=r*(Xr[0].c5+Xr[1].c6);
      (*X).c5.im=r*(Xr[0].c6-Xr[1].c5);
      (*X).c6.re=r*(Xr[1].c6-Xr[0].c5);
      (*X).c6.im=r*(Xr[1].c5+Xr[0].c6);

      (*X).c7.re=r*(Xr[0].c7+Xr[1].c8);
      (*X).c7.im=r*(Xr[0].c8-Xr[1].c7);
      (*X).c8.re=r*(Xr[1].c8-Xr[0].c7);
      (*X).c8.im=r*(Xr[1].c7+Xr[0].c8);
   }
}


double norm_square_calg(int vol,int icom,sl3_alg_dble *X)
{
   double sm;
   complex_dble z;
   sl3_alg_dble *Xm;

   if (init==0)
      set_ids();

   reset_hsum(idsmr);
   Xm=X+vol;

   for (;X<Xm;X++)
   {
      z.re=(*X).c1.re+(*X).c2.re;
      z.im=(*X).c1.im+(*X).c2.im;
      sm=z.re*z.re+z.im*z.im;

      sm+=(*X).c1.re*(*X).c1.re+(*X).c1.im*(*X).c1.im;
      sm+=(*X).c2.re*(*X).c2.re+(*X).c2.im*(*X).c2.im;
      sm+=(*X).c3.re*(*X).c3.re+(*X).c3.im*(*X).c3.im;
      sm+=(*X).c4.re*(*X).c4.re+(*X).c4.im*(*X).c4.im;
      sm+=(*X).c5.re*(*X).c5.re+(*X).c5.im*(*X).c5.im;
      sm+=(*X).c6.re*(*X).c6.re+(*X).c6.im*(*X).c6.im;
      sm+=(*X).c7.re*(*X).c7.re+(*X).c7.im*(*X).c7.im;
      sm+=(*X).c8.re*(*X).c8.re+(*X).c8.im*(*X).c8.im;

      add_to_hsum(idsmr,&sm);
   }

   if (icom==1)
      global_hsum(idsmr,&sm);
   else
      local_hsum(idsmr,&sm);

   return 2.0*sm;
}


complex_dble scalar_prod_calg(int vol,int icom,sl3_alg_dble *X,sl3_alg_dble *Y)
{
   complex_dble sm,zx,zy;
   sl3_alg_dble *Xm;

   if (init==0)
      set_ids();

   reset_hsum(idsmz);
   Xm=X+vol;

   for (;X<Xm;X++)
   {
      zx.re=(*X).c1.re+(*X).c2.re;
      zx.im=(*X).c1.im+(*X).c2.im;
      zy.re=(*Y).c1.re+(*Y).c2.re;
      zy.im=(*Y).c1.im+(*Y).c2.im;

      sm.re=zx.re*zy.re+zx.im*zy.im;
      sm.im=zx.re*zy.im-zx.im*zy.re;

      sm.re+=(*X).c1.re*(*Y).c1.re+(*X).c1.im*(*Y).c1.im;
      sm.im+=(*X).c1.re*(*Y).c1.im-(*X).c1.im*(*Y).c1.re;

      sm.re+=(*X).c2.re*(*Y).c2.re+(*X).c2.im*(*Y).c2.im;
      sm.im+=(*X).c2.re*(*Y).c2.im-(*X).c2.im*(*Y).c2.re;

      sm.re+=(*X).c3.re*(*Y).c3.re+(*X).c3.im*(*Y).c3.im;
      sm.im+=(*X).c3.re*(*Y).c3.im-(*X).c3.im*(*Y).c3.re;

      sm.re+=(*X).c4.re*(*Y).c4.re+(*X).c4.im*(*Y).c4.im;
      sm.im+=(*X).c4.re*(*Y).c4.im-(*X).c4.im*(*Y).c4.re;

      sm.re+=(*X).c5.re*(*Y).c5.re+(*X).c5.im*(*Y).c5.im;
      sm.im+=(*X).c5.re*(*Y).c5.im-(*X).c5.im*(*Y).c5.re;

      sm.re+=(*X).c6.re*(*Y).c6.re+(*X).c6.im*(*Y).c6.im;
      sm.im+=(*X).c6.re*(*Y).c6.im-(*X).c6.im*(*Y).c6.re;

      sm.re+=(*X).c7.re*(*Y).c7.re+(*X).c7.im*(*Y).c7.im;
      sm.im+=(*X).c7.re*(*Y).c7.im-(*X).c7.im*(*Y).c7.re;

      sm.re+=(*X).c8.re*(*Y).c8.re+(*X).c8.im*(*Y).c8.im;
      sm.im+=(*X).c8.re*(*Y).c8.im-(*X).c8.im*(*Y).c8.re;

      add_to_hsum(idsmz,(double*)(&sm));
      Y+=1;
   }

   if (icom==1)
      global_hsum(idsmz,(double*)(&sm));
   else
      local_hsum(idsmz,(double*)(&sm));

   sm.re*=2.0;
   sm.im*=2.0;

   return sm;
}


void set_calg2zero(int vol,sl3_alg_dble *X)
{
   sl3_alg_dble *Xm;

   Xm=X+vol;

   for (;X<Xm;X++)
      sl3_zero(X);
}


void assign_calg2calg(int vol,sl3_alg_dble *X,sl3_alg_dble *Y)
{
   sl3_alg_dble *Xm;

   Xm=X+vol;

   for (;X<Xm;X++)
   {
      (*Y)=(*X);
      Y+=1;
   }
}


void assign_calgdag2calg(int vol,sl3_alg_dble *X,sl3_alg_dble *Y)
{
   complex_dble z;
   sl3_alg_dble *Xm;

   Xm=X+vol;

   for (;X<Xm;X++)
   {
      (*Y).c1.re=(*X).c1.re;
      (*Y).c1.im=-(*X).c1.im;
      (*Y).c2.re=(*X).c2.re;
      (*Y).c2.im=-(*X).c2.im;

      z.re=(*X).c3.re;
      z.im=(*X).c3.im;
      (*Y).c3.re=(*X).c4.re;
      (*Y).c3.im=-(*X).c4.im;
      (*Y).c4.re=z.re;
      (*Y).c4.im=-z.im;

      z.re=(*X).c5.re;
      z.im=(*X).c5.im;
      (*Y).c5.re=(*X).c6.re;
      (*Y).c5.im=-(*X).c6.im;
      (*Y).c6.re=z.re;
      (*Y).c6.im=-z.im;

      z.re=(*X).c7.re;
      z.im=(*X).c7.im;
      (*Y).c7.re=(*X).c8.re;
      (*Y).c7.im=-(*X).c8.im;
      (*Y).c8.re=z.re;
      (*Y).c8.im=-z.im;

      Y+=1;
   }
}


void add_calg(int vol,sl3_alg_dble *X,sl3_alg_dble *Y)
{
   sl3_alg_dble *Xm;

   Xm=X+vol;

   for (;X<Xm;X++)
   {
      sl3_add(X,Y);
      Y+=1;
   }
}


void sub_calg(int vol,sl3_alg_dble *X,sl3_alg_dble *Y)
{
   sl3_alg_dble *Xm;

   Xm=X+vol;

   for (;X<Xm;X++)
   {
      sl3_sub(X,Y);
      Y+=1;
   }
}


void mulr_calg(int vol,double r,sl3_alg_dble *X,sl3_alg_dble *Y)
{
   sl3_alg_dble *Xm;

   Xm=X+vol;

   for (;X<Xm;X++)
   {
      sl3_mulr(&r,X,Y);
      Y+=1;
   }
}


void mulr_add_calg(int vol,sl3_alg_dble *X,sl3_alg_dble *Y,double r)
{
   sl3_alg_dble *Xm;

   Xm=X+vol;

   for (;X<Xm;X++)
   {
      sl3_mulr_add(&r,Y,X);
      Y+=1;
   }
}


void mulc_add_calg(int vol,sl3_alg_dble *X,sl3_alg_dble *Y,complex_dble z)
{
   complex_dble w;
   sl3_alg_dble *Xm;

   Xm=X+vol;

   for (;X<Xm;X++)
   {
      w.re=z.re*(*Y).c1.re-z.im*(*Y).c1.im;
      w.im=z.re*(*Y).c1.im+z.im*(*Y).c1.re;
      (*X).c1.re+=w.re;
      (*X).c1.im+=w.im;

      w.re=z.re*(*Y).c2.re-z.im*(*Y).c2.im;
      w.im=z.re*(*Y).c2.im+z.im*(*Y).c2.re;
      (*X).c2.re+=w.re;
      (*X).c2.im+=w.im;

      w.re=z.re*(*Y).c3.re-z.im*(*Y).c3.im;
      w.im=z.re*(*Y).c3.im+z.im*(*Y).c3.re;
      (*X).c3.re+=w.re;
      (*X).c3.im+=w.im;

      w.re=z.re*(*Y).c4.re-z.im*(*Y).c4.im;
      w.im=z.re*(*Y).c4.im+z.im*(*Y).c4.re;
      (*X).c4.re+=w.re;
      (*X).c4.im+=w.im;

      w.re=z.re*(*Y).c5.re-z.im*(*Y).c5.im;
      w.im=z.re*(*Y).c5.im+z.im*(*Y).c5.re;
      (*X).c5.re+=w.re;
      (*X).c5.im+=w.im;

      w.re=z.re*(*Y).c6.re-z.im*(*Y).c6.im;
      w.im=z.re*(*Y).c6.im+z.im*(*Y).c6.re;
      (*X).c6.re+=w.re;
      (*X).c6.im+=w.im;

      w.re=z.re*(*Y).c7.re-z.im*(*Y).c7.im;
      w.im=z.re*(*Y).c7.im+z.im*(*Y).c7.re;
      (*X).c7.re+=w.re;
      (*X).c7.im+=w.im;

      w.re=z.re*(*Y).c8.re-z.im*(*Y).c8.im;
      w.im=z.re*(*Y).c8.im+z.im*(*Y).c8.re;
      (*X).c8.re+=w.re;
      (*X).c8.im+=w.im;

      Y+=1;
   }
}
