
/*******************************************************************************
*
* File rdlist.c
*
* Copyright (C) 2015 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Utility programs for lists of rooted labeled tree diagrams in the
* recursive representation.
*
*   rlist_t *alloc_rlist(int nmax)
*     Creates a new diagram list, with initially space for nmax diagrams.
*     The program returns the pointer to the new list.
*
*   void free_rlist(rlist_t *rl)
*     Frees the arrays in the diagram list rl and then the structure
*     itself. The diagrams pointed to by the diagram array and the
*     coefficients pointed to by the coefficient array are not freed.
*     The program does nothing if rl=NULL.
*
*   void add_rtd(rtree_t *rd,cnum_t *cn,rlist_t *rl)
*     Adds the rooted labeled tree diagram rd and its coefficient cn to
*     the diagram list rl.
*
* The structures of type rlist_t have the following elements:
*
*  n                     Current number of diagrams included in the
*                        list.
*
*  nmax                  Current maximal number of diagrams that can
*                        be included in the list. nmax and the arrays in
*                        the list are adjusted dynamically by add_rtd().
*
*  rta[i]                Pointer to the rooted labeled tree diagram
*                        number i (i=0,..,n-1).
*
*  cna[i]                Pointer to the coefficient (of type cnum_t) of
*                        the i'th diagram.
*
* In the recursive representation, rooted labeled tree diagrams are described
* by structures of type rtree_t [ispt/rtree.c]. Unused elements of the diagram
* and coefficient array are set to NULL.
*
* The type cnum_t is defined in utils.h and programs manipulating numbers
* of this type are collected in utils/cnum.c.
*
* Note that the array addresses (*rl).rta and (*rl).cna may change when a
* diagram is added to the diagram list rl.
*
* The programs in this module do not perform any communications and can be
* locally called.
*
*******************************************************************************/

#define RDLIST_C

#include <stdlib.h>
#include <stdio.h>
#include <limits.h>
#include "utils.h"
#include "dmlist.h"


static rtree_t **new_rta(int nmax,int nold,rtree_t **rtold)
{
   int i;
   rtree_t **rta;

   rta=malloc(nmax*sizeof(*rta));
   error_loc(rta==NULL,1,"new_rta [rdlist.c]",
             "Unable to allocate diagram array");

   for (i=0;i<nold;i++)
      rta[i]=rtold[i];

   for (i=nold;i<nmax;i++)
      rta[i]=NULL;

   if (nold>0)
      free(rtold);

   return rta;
}


static cnum_t **new_cna(int nmax,int nold,cnum_t **cnold)
{
   int i;
   cnum_t **cna;

   cna=malloc(nmax*sizeof(*cna));
   error_loc(cna==NULL,1,"new_cna [rdlist.c]",
             "Unable to allocate coefficient array");

   for (i=0;i<nold;i++)
      cna[i]=cnold[i];

   for (i=nold;i<nmax;i++)
      cna[i]=NULL;

   if (nold>0)
      free(cnold);

   return cna;
}


rlist_t *alloc_rlist(int nmax)
{
   rlist_t *rl;

   error_loc(nmax<1,1,"alloc_rlist [rdlist.c]",
             "Argument nmax is out of range");

   rl=malloc(1*sizeof(*rl));
   error_loc(rl==NULL,1,"alloc_rlist [rdlist.c]",
             "Unable to allocate diagram list");

   (*rl).n=0;
   (*rl).nmax=nmax;
   (*rl).rta=new_rta(nmax,0,NULL);
   (*rl).cna=new_cna(nmax,0,NULL);

   return rl;
}


void free_rlist(rlist_t *rl)
{
   if (rl!=NULL)
   {
      free((*rl).rta);
      free((*rl).cna);
      free(rl);
   }
}


void add_rtd(rtree_t *rd,cnum_t *cn,rlist_t *rl)
{
   int m,n,nmax;

   n=(*rl).n;
   nmax=(*rl).nmax;

   if (n==nmax)
   {
      if (nmax<=(INT_MAX/2))
         m=2*nmax;
      else
      {
         error_loc(nmax==INT_MAX,1,"add_rtd [rdlist.c]",
                   "Unable to extend diagram list");
         m=INT_MAX;
      }

      (*rl).nmax=m;
      (*rl).rta=new_rta(m,nmax,(*rl).rta);
      (*rl).cna=new_cna(m,nmax,(*rl).cna);
   }

   (*rl).n=n+1;
   (*rl).rta[n]=rd;
   (*rl).cna[n]=cn;
}
