
/*******************************************************************************
*
* File lnlist.c
*
* Copyright (C) 2014 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Generation of line lists.
*
*   long init_lnlist(int m,int *nv,int *nl,int **v,int **ne)
*     Initializes the line-list generator and returns the number of
*     lists of the type specified by the arguments m,..,ne (see the
*     notes).
*
*   long next_lnlist(int (*p)[2],long *nc)
*     Assigns the next line list of the type specified by init_lnlist()
*     to the array p and returns the current list count (see the notes).
*     The number of Wick contractions of the external lines leading to
*     the line list p is assigned to nc. The generator wraps around when
*     all possible line lists have been generated.
*
*   void stop_lnlist(void)
*     Resets the line list generator to its uninitialized state. Note
*     that a running generator must be stopped by calling this program
*     before it can be re-initialized.
*
* A line list is an array of pairs
*
*  (p[0][0],p[0][1]),..,(p[n-1][0],p[n-1][1])
*
* of vertex indices, which are interpreted as the endpoints of a set of
* boson lines (n lines in this example). Vertices are abstractly defined
* by an index and the number of external boson lines attached to them.
* Each vertex may not occur more often in the list than the number of its
* external lines.
*
* The program next_lnlist() simultaneously cycles through the line lists
* associated to one or more sets of vertices. These must be specified when
* the generator is initialized together with a few further parameters:
*
*  m         Number of vertex sets.
*
*  nv        Array of the numbers nv[i], i=0,..,m-1, of vertices in
*            i'th vertex set.
*
*  nl        Array of the numbers nl[i] of the lines in the line list
*            associated to the i'th vertex set.
*
*  v         Array of the indices v[i][j], j=0,..,nv[i]-1, of the j'th
*            vertex in the i'th vertex set.
*
*  ne        Array of the numbers ne[i][j] of the external lines of
*            the j'th vertex in the i'th vertex set.
*
* These must satisfy the constraints
*
*  m>0, nv[i]>0, nl[i]>=0, v[i][j]!=v[i][l] if j!=l, ne[i][j]>=0.
*
* An error occurs if one of these is violated. The chosen values of the
* parameters must be such that the total number of line lists is less than
* or equal to LONG_MAX (on x86-64 machines, this is 2^63-1 and thus about
* 9.2x10^18).
*
* Any two line lists are considered to be the same, if they coincide after
* a reordering of the lines and their endpoints. The program next_lnlist()
* cycles through all different line lists only. The first nl[0] lines of
* the line list returned by the program refer to the first vertex set, the
* next nl[1] lines to the second vertex set, and so on.
*
* The programs in this module do not perform any communications and can be
* locally called.
*
*******************************************************************************/

#define LNLIST_C

#include <stdlib.h>
#include <stdio.h>
#include <limits.h>
#include "utils.h"
#include "permutations.h"

static int ma=0,nva=0,nla=0,nma=0,init=0;
static int ms,*nvs,**vs,**nes;
static int *nls,(**lns)[2];
static int *ivs,*ils;
static long np,npm;


static void alloc_arrays(int m,int *nv,int *nl)
{
   int i,nvt,nlt,nvm,nlm;

   nvt=0;
   nlt=0;
   nvm=0;
   nlm=0;

   for (i=0;i<m;i++)
   {
      nvt+=nv[i];
      nlt+=nl[i];

      if (nv[i]>nvm)
         nvm=nv[i];
      if (nl[i]>nlm)
         nlm=nl[i];
   }

   if ((m>ma)||(nvt>nva)||(nlt>nla)||((nvm+nlm)>nma))
   {
      if (nva>0)
         free(vs[0]);
      if (nla>0)
         free(lns[0]);
      if (ma>0)
      {
         free(nvs);
         free(vs);
         free(lns);
      }

      if (m>ma)
         ma=m;
      if (nvt>nva)
         nva=nvt;
      if (nlt>nla)
         nla=nlt;
      if ((nvm+nlm)>nma)
         nma=nvm+nlm;

      nvs=malloc((2*ma+nma)*sizeof(*nvs));
      vs=malloc(2*ma*sizeof(*vs));
      lns=malloc(ma*sizeof(*lns));

      error_loc((nvs==NULL)||(vs==NULL)||(lns==NULL),1,
                "alloc_arrays [lnlist.c]",
                "Unable to allocate auxiliary arrays");

      vs[0]=malloc(2*nva*sizeof(**vs));

      if (nla>0)
         lns[0]=malloc(nla*sizeof(**lns));
      else
         lns[0]=NULL;

      error_loc((vs[0]==NULL)||((nla>0)&&(lns[0]==NULL)),2,
                "alloc_arrays [lnlist.c]",
                "Unable to allocate auxiliary arrays");
   }

   nls=nvs+m;
   ivs=nls+m;
   ils=ivs+nvm;

   nes=vs+m;
   nes[0]=vs[0]+nvt;

   for (i=1;i<m;i++)
   {
      vs[i]=vs[i-1]+nv[i-1];
      nes[i]=nes[i-1]+nv[i-1];
      lns[i]=lns[i-1]+nl[i-1];
   }
}


static void reset_lns(int i)
{
   int j,k;

   k=0;

   for (j=0;j<nls[i];j++)
   {
      while (nes[i][k]==0)
         k+=1;

      nes[i][k]-=1;
      lns[i][j][0]=k;

      while (nes[i][k]==0)
         k+=1;

      nes[i][k]-=1;
      lns[i][j][1]=k;
   }
}


static void set_arrays(int m,int *nv,int *nl,int **v,int **ne)
{
   int i,j;

   ms=m;

   for (i=0;i<m;i++)
   {
      nvs[i]=nv[i];
      nls[i]=nl[i];

      for (j=0;j<nv[i];j++)
      {
         vs[i][j]=v[i][j];
         nes[i][j]=ne[i][j];
      }

      reset_lns(i);
   }
}


static void assign_lns(int (*p)[2])
{
   int i,j;

   for (i=0;i<ms;i++)
   {
      for (j=0;j<nls[i];j++)
      {
         p[0][0]=vs[i][lns[i][j][0]];
         p[0][1]=vs[i][lns[i][j][1]];
         p+=1;
      }
   }
}


static int set_later(int i,int j)
{
   int k0,k1,k,l;
   int n0,n1;

   if (j==(nls[i]-1))
      return 1;

   k0=lns[i][j][0];
   k1=lns[i][j][1];
   n1=0;

   for (k=k0+1;k<nvs[i];k++)
      n1+=nes[i][k];

   if (k1==k0)
   {
      if ((nes[i][k0]&0x1)&&(n1>0))
      {
         n0=nes[i][k0]/2+1;
         n1-=1;
      }
      else
         n0=nes[i][k0]/2;
   }
   else
   {
      n0=0;

      for (k=k1;k<nvs[i];k++)
         n0+=nes[i][k];

      if (n0>nes[i][k0])
         n0=nes[i][k0];

      n1-=n0;
   }

   if ((n0+n1/2)<(nls[i]-j-1))
   {
      nes[i][k0]+=1;
      nes[i][k1]+=1;

      return 0;
   }
   else
   {
      j+=1;

      if (k1==k0)
      {
         for (l=0;(l<n0)&&(j<nls[i]);l++)
         {
            if (nes[i][k0]>1)
            {
               nes[i][k0]-=2;
               lns[i][j][0]=k0;
               lns[i][j][1]=k0;
            }
            else
            {
               k=k0+1;

               while (nes[i][k]==0)
                  k+=1;

               nes[i][k0]-=1;
               nes[i][k]-=1;
               lns[i][j][0]=k0;
               lns[i][j][1]=k;
            }

            j+=1;
         }
      }
      else
      {
         k=k1;

         for (l=0;(l<n0)&&(j<nls[i]);l++)
         {
            while (nes[i][k]==0)
               k+=1;

            nes[i][k0]-=1;
            nes[i][k]-=1;
            lns[i][j][0]=k0;
            lns[i][j][1]=k;
            j+=1;
         }
      }

      k=k0+1;

      while (j<nls[i])
      {
         while (nes[i][k]==0)
            k+=1;

         nes[i][k]-=1;
         lns[i][j][0]=k;

         while (nes[i][k]==0)
            k+=1;

         nes[i][k]-=1;
         lns[i][j][1]=k;
         j+=1;
      }

      return 1;
   }
}


static int fnd_next(int i,int j)
{
   int k0,k1,k2;

   k1=lns[i][j][1];
   nes[i][k1]+=1;

   for (k2=k1+1;k2<nvs[i];k2++)
   {
      if (nes[i][k2]>0)
      {
         nes[i][k2]-=1;
         lns[i][j][1]=k2;
         return set_later(i,j);
      }
   }

   k0=lns[i][j][0];
   nes[i][k0]+=1;

   for (k1=k0+1;k1<nvs[i];k1++)
   {
      if (nes[i][k1]>0)
      {
         nes[i][k1]-=1;
         lns[i][j][0]=k1;

         for (k2=k1;k2<nvs[i];k2++)
         {
            if (nes[i][k2]>0)
            {
               nes[i][k2]-=1;
               lns[i][j][1]=k2;
               return set_later(i,j);
            }
         }

         nes[i][k1]+=1;
         return 0;
      }
   }

   return 0;
}


static int cycle(int i,int j)
{
   if (fnd_next(i,j)==1)
      return 1;
   else if (j>0)
      return cycle(i,j-1);
   else
   {
      reset_lns(i);

      if (i<(ms-1))
         return cycle(i+1,nls[i+1]-1);
      else
         return 0;
   }
}


static long set_npm(void)
{
   int ic;

   ic=1;
   npm=0L;

   while (ic==1)
   {
      if (npm<LONG_MAX)
         npm+=1L;
      else
         error_loc(1,1,"set_npm [lnlist.c]",
                   "Number of line lists is out of range");

      ic=cycle(0,nls[0]-1);
   }

   return npm;
}


static void set_ivs(int i)
{
   int j,k;

   for (j=0;j<nvs[i];j++)
      ivs[j]=0;

   for (j=0;j<nls[i];j++)
   {
      ivs[lns[i][j][0]]+=1;
      ivs[lns[i][j][1]]+=1;
      ils[j]=1;

      for (k=0;k<j;k++)
      {
         if (ils[k]!=0)
         {
            if ((lns[i][j][0]==lns[i][k][0])&&
                (lns[i][j][1]==lns[i][k][1]))
            {
               ils[k]+=1;
               ils[j]=0;
               break;
            }
         }
      }
   }
}


static long nctr(void)
{
   int i,j,k,ie;
   long f0,f1,f2;

   ie=0;
   f0=1L;

   for (i=0;i<ms;i++)
   {
      set_ivs(i);
      f1=1L;

      for (j=0;j<nvs[i];j++)
      {
         f2=binomial(nes[i][j]+ivs[j],ivs[j]);

         if (f1<=(LONG_MAX/f2))
            f1*=f2;
         else
            ie=1;

         f2=factorial(ivs[j]);

         if (f1<=(LONG_MAX/f2))
            f1*=f2;
         else
            ie=1;
      }

      for (j=0;j<nls[i];j++)
      {
         if (ils[j]>1)
            f1/=factorial(ils[j]);

         if ((ils[j]>0)&&(lns[i][j][0]==lns[i][j][1]))
         {
            for (k=0;k<ils[j];k++)
               f1/=2L;
         }
      }

      if (f0<=(LONG_MAX/f1))
         f0*=f1;
      else
         ie=1;
   }

   error_loc(ie!=0,1,"nctr [lnlist.c]","Unable to compute contraction count");

   return f0;
}


long init_lnlist(int m,int *nv,int *nl,int **v,int **ne)
{
   int ie,i,j,k,s;

   error_loc(init==1,1,"init_lnlist [lnlist.c]",
             "A running line-list generator cannot be re-initialized");

   ie=(m<1);

   for (i=0;i<m;i++)
   {
      ie|=(nv[i]<=0);
      ie|=(nl[i]<0);

      for (j=1;j<nv[i];j++)
      {
         for (k=0;k<j;k++)
            ie|=(v[i][j]==v[i][k]);
      }

      s=0;

      for (j=0;j<nv[i];j++)
      {
         ie|=(ne[i][j]<0);
         s+=ne[i][j];
      }

      ie|=(s<(2*nl[i]));
   }

   if (ie!=0)
   {
      printf("m = %d, nv = ",m);

      for (i=0;i<m;i++)
      {
         if (i<(m-1))
            printf(" %d,",nv[i]);
         else
            printf(" %d\n",nv[i]);
      }

      printf("nl = ");

      for (i=0;i<m;i++)
      {
         if (i<(m-1))
            printf(" %d,",nl[i]);
         else
            printf(" %d\n",nl[i]);
      }

      for (i=0;i<m;i++)
      {
         printf("v[%d] = ",i);

         for (j=0;j<nv[i];j++)
         {
            if (j<(nv[i]-1))
               printf(" %d,",v[i][j]);
            else
               printf(" %d\n",v[i][j]);
         }
      }

      for (i=0;i<m;i++)
      {
         printf("ne[%d] = ",i);

         for (j=0;j<nv[i];j++)
         {
            if (j<(nv[i]-1))
               printf(" %d,",ne[i][j]);
            else
               printf(" %d\n",ne[i][j]);
         }
      }
   }

   error_loc(ie!=0,1,"init_lnlist [lnlist.c]","Parameters are out of range");

   alloc_arrays(m,nv,nl);
   set_arrays(m,nv,nl,v,ne);
   set_npm();

   np=0L;
   init=1;

   return npm;
}


long next_lnlist(int (*p)[2],long *nc)
{
   int ic;

   error_loc(init!=1,1,"next_lnlist [lnlist.c]",
             "Line-list generator is not initialized");

   assign_lns(p);
   (*nc)=nctr();

   ic=cycle(0,nls[0]-1);
   np+=1L;

   if (ic==1)
      return np;
   else
   {
      np=0L;
      return npm;
   }
}


void stop_lnlist(void)
{
   if (init==1)
      init=0;
}
