
/*******************************************************************************
*
* File permgen.c
*
* Copyright (C) 2014 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Generation of permutations.
*
*   long init_perm(int m,int *n)
*     Initializes the permutation generator and returns the number of
*     permutations of the type specified by the arguments m and n (see
*     the notes).
*
*   long next_perm(int *p)
*     Assigns the next permutation of the type specified by init_perm()
*     to the array p and returns the current permutation count (see the
*     notes). The generator wraps around when all possible permutations
*     have been generated.
*
*   void stop_perm(void)
*     Resets the permutation generator to its uninitialized state. Note
*     that a running generator must be stopped by calling this program
*     before it can be re-initialized.
*
* The program next_perm() permutes the elements of the sequence
*
*   0,1,..,nt-1
*
* of integers in a way that depends on the values of the parameters
*
*   m,n[0],n[1],..,n[m-1]
*
* passed to the initialization program init_perm(). More precisely, the
* total length nt of the sequence is set to n[0]+..+n[m-1]. The generator
* then permutes the first n[0] numbers, the next n[1] numbers, etc., of
* the sequence among themselves. There are
*
*   n[0]!*n[1]!*..*n[m-1]!
*
* such permutations and all of them are obtained by repeatedly calling
* next_perm(). Clearly, the array p passed to the program next_perm() must
* have at least nt elements.
*
* The program next_perm() cycles through the permutations like m nested
* for-loops, where the innermost loop runs over the permutations of the
* first n[0] numbers, the next-to-innermost loop over the permutations of
* the next n[1] numbers, and so on.
*
* The values of m,n[0],..,n[m-1] must satisfy the bounds
*
*  m>0, n[i]>0 for all i=0,..,m-1,
*
* and be such that the number of permutations is less than or equal to
* LONG_MAX (on x86-64 machines, this is 2^63-1 and thus about 9.2x10^18).
* The program init_perm() aborts with an error message if these conditions
* are not satisfied.
*
* The programs in this module do not perform any communications and can be
* locally called.
*
*******************************************************************************/

#define PERMGEN_C

#include <stdlib.h>
#include <stdio.h>
#include <limits.h>
#include "utils.h"
#include "permutations.h"

static int ma=0,nta=0,init=0;
static int **ps,**js,*ns,ms;
static long np,npm;


static void alloc_arrays(int m,int *n)
{
   int k,nt;

   nt=0;

   for (k=0;k<m;k++)
      nt+=n[k];

   if ((m>ma)||(nt>nta))
   {
      if (ma>0)
      {
         free(ps);
         free(ns);
      }

      if (m>ma)
         ma=m;
      if (nt>nta)
         nta=nt;

      ps=malloc(2*ma*sizeof(*ps));
      ns=malloc((ma+2*nta)*sizeof(*ns));

      error_loc((ps==NULL)||(ns==NULL),1,"alloc_arrays [permgen.c]",
                "Unable to allocate auxiliary arrays");
   }

   ms=m;
   js=ps+m;
   ps[0]=ns+m;
   js[0]=ns+m+nt;

   for (k=0;k<m;k++)
   {
      ns[k]=n[k];

      if (k>0)
      {
         ps[k]=ps[k-1]+n[k-1];
         js[k]=js[k-1]+n[k-1];
      }
   }
}


static void set_arrays(void)
{
   int i,k,l;

   i=0;

   for (k=0;k<ms;k++)
   {
      for (l=0;l<ns[k];l++)
      {
         js[k][l]=l;
         ps[k][l]=i;
         i+=1;
      }
   }
}


static void set_npm(void)
{
   int i;
   long f;

   npm=1L;

   for (i=0;i<ms;i++)
   {
      f=factorial(ns[i]);

      if ((LONG_MAX/f)>=npm)
         npm*=f;
      else
         error_loc(1,1,"set_npm [permgen.c]",
                   "Number of permutations is out of range");
   }
}


static void assign_ps(int *p)
{
   int k,l;

   for (k=0;k<ms;k++)
   {
      for (l=0;l<ns[k];l++)
      {
         p[0]=ps[k][l];
         p+=1;
      }
   }
}


static int cycle(int k,int l)
{
   int s,i,j;

   j=js[k][l];
   s=ps[k][j];

   if (j<(ns[k]-1))
   {
      ps[k][j]=ps[k][j+1];
      ps[k][j+1]=s;
      js[k][l]+=1;

      return 1;
   }
   else
   {
      for (i=(ns[k]-1);i>l;i--)
         ps[k][i]=ps[k][i-1];

      ps[k][l]=s;
      js[k][l]=l;

      if (l<(ns[k]-2))
         return cycle(k,l+1);
      else if (k<(ms-1))
         return cycle(k+1,0);
      else
         return 0;
   }
}


long init_perm(int m,int *n)
{
   int k;

   error_loc(init==1,1,"init_perm [permgen.c]",
             "A running permutation generator cannot be re-initialized");

   error_loc(m<1,1,"init_perm [permgen.c]",
             "There must be at least one group of elements to be permuted");

   for (k=0;k<m;k++)
   {
      if (n[k]<1)
         error_loc(1,1,"init_perm [permgen.c]",
                   "Subsequence length is out of range");
   }

   alloc_arrays(m,n);
   set_arrays();
   set_npm();
   np=0L;
   init=1;

   return npm;
}


long next_perm(int *p)
{
   int ic;

   if (init!=1)
   {
      error_loc(1,1,"next_perm [permgen.c]",
                "Permutation generator is not initialized");
   }

   assign_ps(p);
   ic=cycle(0,0);
   np+=1L;

   if (ic==1)
      return np;
   else
   {
      np=0L;
      return npm;
   }
}


void stop_perm(void)
{
   if (init==1)
      init=0;
}
