
/*******************************************************************************
*
* File wickgen.c
*
* Copyright (C) 2014 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Generation of Wick contractions.
*
*   long init_wick(int m,int nb,int *n,int **a)
*     Initializes the contraction generator and returns the number of
*     contractions of the type specified by the arguments (see the notes).
*
*   long next_wick(int (*p)[2])
*     Assigns the next contraction of the type specified by init_wick()
*     to the array p and returns the current contraction count (see the
*     notes). The generator wraps around when all possible contractions
*     have been generated.
*
*   void stop_wick(void)
*     Resets the contraction generator to its uninitialized state. Note
*     that a running generator must be stopped by calling this program
*     before it can be re-initialized.
*
* The program next_wick() pairs the elements of an array
*
*   a[k][l],    k=0,..,m-1,   l=0,..,n[k]-1,
*
* of integers as specified in the following lines. Only the elements within
* each of the one-dimensional arrays a[k] are paired. The pairing is "bosonic"
* if 0<=k<nb and "fermionic" otherwise. In the first case, all elements in
* the array a[k] are paired and the elements in each pair are unordered. There
* are (n[k]-1)!! such pairings. "Fermionic" pairings, on the other hand, are
* obtained by pairing the elements in the first half of the array a[k] with
* those in the second half. The pairs are thus of the form (a[k][l1],a[k][l2])
* with l1<(n[k]/2) and l2>=(n[k]/2). There are (n[k]/2)! different pairings of
* this kind.
*
* The arguments of the program init_wick() are:
*
*   m           Number of arrays a[k].
*
*   nb          Number of arrays whose elements are to be paired
*               in the "bosonic" way. The elements in the other m-nb
*               arrays are paired in the "fermionic" way.
*
*   n           Address of the array containing the numbers of elements of
*               the arrays a[0],..,a[m-1].
*
*   a           Address of the two-dimensional array containing the elements
*               that are to be paired.
*
* The arguments must satisfy the constraints
*
*   m>0,  0<=nb<=m,  n[k]>0,  n[k]=even (k=0,..,m-1),
*
* and be such that the number of pairings is less than or equal to LONG_MAX
* (on x86-64 machines, this is 2^63-1 and thus about 9.2x10^18). The program
* init_wick() aborts with an error message if these conditions are not met.
* There are no constraints on the array elements a[k][l]. The pairing of the
* array elements is in fact independent of their values.
*
* The pairings generated by next_wick() are delivered in the form of a list
*
*   (p[0][0],p[0][1]),(p[1][0],p[1][1]),...
*
* of element pairs, where the first n[0]/2 pairs are built from the elements
* of a[0], the next n[1]/2 pairs from those of a[1], and so on. The program
* generates all possible pairings of the type specified by the data passed
* to the initialization program. The counting starts from 1.
*
* The programs in this module do not perform any communications and can be
* locally called.
*
*******************************************************************************/

#define WICKGEN_C

#include <stdlib.h>
#include <stdio.h>
#include <limits.h>
#include "utils.h"
#include "permutations.h"

static int ma=0,nta=0,init=0;
static int **ps,**js,*ns,ms,nbs;
static long np,npm;


static void alloc_arrays(int m,int nb,int *n)
{
   int k,nt;

   nt=0;

   for (k=0;k<m;k++)
      nt+=n[k];

   if ((m>ma)||(nt>nta))
   {
      if (ma>0)
      {
         free(ps);
         free(ns);
      }

      if (m>ma)
         ma=m;
      if (nt>nta)
         nta=nt;

      ps=malloc(2*ma*sizeof(*ps));
      ns=malloc((ma+2*nta)*sizeof(*ns));

      error_loc((ps==NULL)||(ns==NULL),1,"alloc_arrays [wickgen.c]",
                "Unable to allocate auxiliary arrays");
   }

   ms=m;
   nbs=nb;

   js=ps+m;
   ps[0]=ns+m;
   js[0]=ns+m+nt;

   for (k=1;k<m;k++)
   {
      ps[k]=ps[k-1]+n[k-1];
      js[k]=js[k-1]+n[k-1];
   }
}


static void set_arrays(int *n,int **a)
{
   int k,l;

   for (k=0;k<ms;k++)
   {
      ns[k]=n[k];

      for (l=0;l<n[k];l++)
      {
         js[k][l]=l;
         ps[k][l]=a[k][l];
      }
   }
}


static void set_npm(void)
{
   int k;
   long f;

   npm=1L;

   for (k=0;k<ms;k++)
   {
      if (k<nbs)
         f=factorial2(ns[k]-1);
      else
         f=factorial(ns[k]/2);

      if ((LONG_MAX/f)>=npm)
         npm*=f;
      else
         error_loc(1,1,"set_npm [wickgen.c]",
                   "Number of contractions is out of range");
   }
}


static void assign_ps(int (*p)[2])
{
   int k,l,n;

   for (k=0;k<ms;k++)
   {
      if (k<nbs)
      {
         for (l=0;l<ns[k];l+=2)
         {
            p[0][0]=ps[k][l];
            p[0][1]=ps[k][l+1];
            p+=1;
         }
      }
      else
      {
         n=ns[k]/2;

         for (l=0;l<n;l++)
         {
            p[0][0]=ps[k][l];
            p[0][1]=ps[k][l+n];
            p+=1;
         }
      }
   }
}


static int cycle(int k,int l)
{
   int s,i,j;

   if (k<nbs)
   {
      j=js[k][l];
      s=ps[k][j];

      if (j<(ns[k]-1))
      {
         ps[k][j]=ps[k][l];
         ps[k][l]=ps[k][j+1];
         ps[k][j+1]=s;
         js[k][l]+=1;

         return 1;
      }
      else
      {
         ps[k][j]=ps[k][l];
         ps[k][l]=s;
         js[k][l]=l;

         if (l>2)
            return cycle(k,l-2);
         else if (k<(ms-1))
         {
            if (k<(nbs-1))
               return cycle(k+1,ns[k+1]-1);
            else
               return cycle(k+1,ns[k+1]/2);
         }
         else
            return 0;
      }
   }
   else
   {
      j=js[k][l];
      s=ps[k][j];

      if (j<(ns[k]-1))
      {
         ps[k][j]=ps[k][j+1];
         ps[k][j+1]=s;
         js[k][l]+=1;

         return 1;
      }
      else
      {
         for (i=(ns[k]-1);i>l;i--)
            ps[k][i]=ps[k][i-1];

         ps[k][l]=s;
         js[k][l]=l;

         if (l<(ns[k]-2))
            return cycle(k,l+1);
         else if (k<(ms-1))
            return cycle(k+1,ns[k+1]/2);
         else
            return 0;
      }
   }
}


long init_wick(int m,int nb,int *n,int **a)
{
   int k;

   error_loc(init==1,1,"init_wick [wickgen.c]",
             "A running Wick generator cannot be re-initialized");

   error_loc((m<1)||(nb<0)||(nb>m),1,"init_wick [wickgen.c]",
             "Improper parameters m,nb");

   for (k=0;k<m;k++)
   {
      if ((n[k]<1)||((n[k]%2)!=0))
         error_loc(1,1,"init_wick [wickgen.c]",
                   "Array lengths must be positive and even");
   }

   alloc_arrays(m,nb,n);
   set_arrays(n,a);
   set_npm();
   np=0L;
   init=1;

   return npm;
}


long next_wick(int (*p)[2])
{
   int ic;

   if (init!=1)
   {
      error_loc(1,1,"next_wick [wickgen.c]",
                "Wick generator is not initialized");
   }

   assign_ps(p);

   if (nbs>0)
      ic=cycle(0,ns[0]-1);
   else
      ic=cycle(0,ns[0]/2);

   np+=1L;

   if (ic==1)
      return np;
   else
   {
      np=0L;
      return npm;
   }
}


void stop_wick(void)
{
   if (init==1)
      init=0;
}
