
/*******************************************************************************
*
* File ispt_parms.c
*
* Copyright (C) 2015 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Specification of the ISPT parameters.
*
*   ispt_parms_t set_ispt_parms(int order,int iroot,int *opv,double **pv)
*     The parameters are
*
*       order          Maximal order in the coupling. Must be greater
*                      or equal to 1.
*
*       iroot          Index of the root-line selection algorithm. Must
*                      be 0 or 1 (see select_root() [ltree_gen.c]).
*
*       opv            Order opv[i] up to which the expansion of the
*                      parameter number i is known, where i runs over
*                      the parameter types listed in pmdb.h.
*
*       pv             Values pv[i][j] of the parameter number i at
*                      order j in the coupling (j=0,..,opv[i])
*
*     The return value is a structure that contains these parameters.
*     The arrays opv and pv in the structure have size (int)(PM_TYPES)
*     and (int)(PM_TYPES) x (order+1), respectively.
*
*   ispt_parms_t ispt_parms(void)
*     Returns the current ispt parameters in a statically allocated
*     structure that contains the above parameters.
*
*   void print_ispt_parms(void)
*     Prints the ispt parameters to stdout on MPI process 0. Only the
*     non-zero values of the higher-order coefficients are printed.
*
*   void write_ispt_parms(FILE *fdat)
*     Writes the ispt parameters to the file fdat on MPI process 0,
*     including any unused array elements.
*
*   void check_ispt_parms(FILE *fdat)
*     Compares the ispt parameters with the values stored on the file
*     fdat on MPI process 0, assuming the latter were written to the
*     file by the program write_ispt_parms().
*
* To ensure the consistency of the data base, the parameters must be set
* simultaneously on all processes. The program set_ispt_parms() may be
* called at most once and only after the theory parameters have been set
* by set_th_parms() [pmdb/th_parms.c].
*
* The parameters that can be specified are of type pm_type_t (see pmdb.h).
* Their meaning is
*
*  C0            Coefficient of the plaquette term in the gauge
*                action (see doc/gfnorm.pdf).
*
*  LAMBDA        Bare gauge parameter.
*
*  MASS          Bare quark mass.
*
*  CSW           Coefficient of the Sheikholeslami-Wohlert term.
*
*  CG,CGP        Boundary improvement coefficients c_G and c_G'.
*
*  CF,CFP        Boundary improvement coefficients c_F and c_F'.
*
* The program set_ispt_parms() sets any unused elements of the arrays
* opv and pv to default values. In particular, if the parameter number
* i is unused, opv[i]=order and pv[i][j]=0.0 for all j=0,..,order. Any
* other unused elements of the array pv are set to 0.0.
*
*******************************************************************************/

#define ISPT_PARMS_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "utils.h"
#include "pmdb.h"
#include "global.h"

#define PMAX (int)(PM_TYPES)

static int ispt_flag=0;
static ispt_parms_t pms={0,0,{0},{NULL}};
static char *pmn[(int)(PM_TYPES)]={"c0","lambda","m",
                                   "csw","cG","cG'","cF","cF'"};


static void alloc_pms(int order)
{
   int i,*opv;
   double **pv,*p;

   p=malloc(PMAX*(order+1)*sizeof(*p));
   error(p==NULL,1,"alloc_pms [ispt_parms.c]",
         "Unable to allocate parameter array");

   for (i=0;i<(PMAX*(order+1));i++)
      p[i]=0.0;

   opv=pms.opv;
   pv=pms.pv;

   for (i=0;i<PMAX;i++)
   {
      opv[i]=order;
      pv[i]=p;
      p+=(order+1);
   }
}


ispt_parms_t set_ispt_parms(int order,int iroot,int *opv,double **pv)
{
   int i,j,ie,is,iprms[2];
   double dprms[1];
   th_parms_t thp;
   pm_type_t pt;

   ie=0;
   thp=th_parms();

   error(thp.th==THEORIES,1,"set_ispt_parms [ispt_parms.c]",
         "Theory parameters are not set");

   error(ispt_flag!=0,1,"set_ispt_parms [ispt_parms.c]",
         "Attempt to reset the ISPT parameters");

   if (NPROC>1)
   {
      iprms[0]=order;
      iprms[1]=iroot;
      MPI_Bcast(iprms,2,MPI_INT,0,MPI_COMM_WORLD);
      error((iprms[0]!=order)||(iprms[1]!=iroot),1,
            "set_ispt_parms [ispt_parms.c]",
            "Parameters order or iroot are not global");
   }

   error_root(order<1,1,"set_ispt_parms [ispt_parms.c]",
              "Parameter order is out of range");

   error_root((iroot<0)||(iroot>1),1,"set_ispt_parms [ispt_parms.c]",
              "Parameter iroot is out of range");

   if (NPROC>1)
   {
      for (i=0;i<PMAX;i++)
      {
         iprms[0]=opv[i];
         MPI_Bcast(iprms,1,MPI_INT,0,MPI_COMM_WORLD);
         ie|=(iprms[0]!=opv[i]);
      }

      error(ie!=0,1,"set_ispt_parms [ispt_parms.c]",
            "Some parameters opv are not global");
   }

   for (i=0;i<PMAX;i++)
      ie|=((opv[i]<0)||(opv[i]>order));

   error_root(ie!=0,1,"set_ispt_parms [ispt_parms.c]",
              "Some parameters opv are out of range");

   if (NPROC>1)
   {
      for (i=0;i<PMAX;i++)
      {
         for (j=0;j<=opv[i];j++)
         {
            dprms[0]=pv[i][j];
            MPI_Bcast(dprms,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
            ie|=(dprms[0]!=pv[i][j]);
         }
      }

      error(ie!=0,1,"set_ispt_parms [ispt_parms.c]",
            "Some parameters pv are not global");
   }

   error_root((thp.th!=PHI4)&&(pv[(int)(C0)][0]<=0.0),1,
              "set_ispt_parms [ispt_parms.c]",
              "The action coefficient c0 must be positive");

   error_root((thp.th!=PHI4)&&(pv[(int)(LAMBDA)][0]<=0.0),1,
              "set_ispt_parms [ispt_parms.c]",
              "The gauge parameter lambda must be positive");

   alloc_pms(order);
   pms.order=order;
   pms.iroot=iroot;

   for (i=0;i<PMAX;i++)
   {
      pt=(pm_type_t)(i);

      if (thp.th==PHI4)
         is=(pt==MASS);
      else
         is=((pt==C0)||(pt==LAMBDA));

      if ((thp.th==LYM)||(thp.th==LQCD))
         is|=((pt==CG)||(pt==CGP));

      if ((thp.th==QCD)||(thp.th==LQCD))
         is|=(pt==MASS);

      if (thp.th==LQCD)
         is|=((pt==CSW)||(pt==CF)||(pt==CFP));

      if (is)
      {
         pms.opv[i]=opv[i];

         for (j=0;j<=opv[i];j++)
            pms.pv[i][j]=pv[i][j];
      }
   }

   if (thp.th==PHI4)
      pmn[(int)(MASS)]="m^2";

   ispt_flag=1;

   return pms;
}


ispt_parms_t ispt_parms(void)
{
   return pms;
}


static void print_pm(pm_type_t pt)
{
   int n,i,j,*opv;
   double **pv;

   opv=pms.opv;
   pv=pms.pv;
   i=(int)(pt);

   for (j=0;j<=opv[i];j++)
   {
      if ((j==0)||(pv[i][j]!=0.0))
      {
         n=fdigits(pv[i][j]);
         printf("%s[%d] = %.*f\n",pmn[i],j,IMAX(n,1),pv[i][j]);
      }
   }
}


void print_ispt_parms(void)
{
   int my_rank;
   th_parms_t thp;
   theory_t th;

   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      error_root(ispt_flag!=1,1,"print_ispt_parms [ispt_parms.c]",
                 "ISPT parameters are not set");

      printf("ISPT parameters:\n");
      printf("order = %d\n",pms.order);
      printf("iroot = %d\n",pms.iroot);

      thp=th_parms();
      th=thp.th;

      if (th==PHI4)
         print_pm(MASS);
      else
      {
         print_pm(C0);
         print_pm(LAMBDA);

         if ((th==QCD)||(th==LQCD))
            print_pm(MASS);

         if ((th==LYM)||(th==LQCD))
         {
            print_pm(CG);
            print_pm(CGP);

            if (th==LQCD)
            {
               print_pm(CSW);
               print_pm(CF);
               print_pm(CFP);
            }
         }
      }

      printf("\n");
   }
}


void write_ispt_parms(FILE *fdat)
{
   int my_rank,endian;
   int i,j,iw;
   stdint_t istd[2];
   double dstd[1];

   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);
   endian=endianness();

   if (my_rank==0)
   {
      error_root(ispt_flag!=1,1,"write_ispt_parms [ispt_parms.c]",
                 "ISPT parameters are not set");

      istd[0]=(stdint_t)(pms.order);
      istd[1]=(stdint_t)(pms.iroot);

      if (endian==BIG_ENDIAN)
         bswap_int(2,istd);

      iw=fwrite(istd,sizeof(stdint_t),2,fdat);

      for (i=0;i<PMAX;i++)
      {
         istd[0]=(stdint_t)(pms.opv[i]);

         if (endian==BIG_ENDIAN)
            bswap_int(1,istd);

         iw+=fwrite(istd,sizeof(stdint_t),1,fdat);
      }

      for (i=0;i<PMAX;i++)
      {
         for (j=0;j<=pms.order;j++)
         {
            dstd[0]=pms.pv[i][j];

            if (endian==BIG_ENDIAN)
               bswap_double(1,dstd);

            iw+=fwrite(dstd,sizeof(double),1,fdat);
         }
      }

      error_root(iw!=(2+PMAX+PMAX*(pms.order+1)),1,
                 "write_ispt_parms [ispt_parms.c]","Incorrect write count");
   }
}


void check_ispt_parms(FILE *fdat)
{
   int my_rank,endian;
   int i,j,ir,ie;
   stdint_t istd[2];
   double dstd[1];

   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);
   endian=endianness();

   if (my_rank==0)
   {
      error_root(ispt_flag!=1,1,"check_ispt_parms [ispt_parms.c]",
                 "ISPT parameters are not set");

      ir=fread(istd,sizeof(stdint_t),2,fdat);

      if (endian==BIG_ENDIAN)
         bswap_int(2,istd);

      ie=(istd[0]!=(stdint_t)(pms.order));
      ie|=(istd[1]!=(stdint_t)(pms.iroot));

      for (i=0;i<PMAX;i++)
      {
         ir+=fread(istd,sizeof(stdint_t),1,fdat);

         if (endian==BIG_ENDIAN)
            bswap_int(1,istd);

         ie|=(istd[0]!=(stdint_t)(pms.opv[i]));
      }

      for (i=0;i<PMAX;i++)
      {
         for (j=0;j<=pms.order;j++)
         {
            ir+=fread(dstd,sizeof(double),1,fdat);

            if (endian==BIG_ENDIAN)
               bswap_double(1,dstd);

            ie|=(dstd[0]!=pms.pv[i][j]);
         }
      }

      error_root(ir!=(2+PMAX+PMAX*(pms.order+1)),1,
                 "check_ispt_parms [ispt_parms.c]","Incorrect read count");

      error_root(ie!=0,1,"check_ispt_parms [ispt_parms.c]",
                 "Parameters do not match");
   }
}
