
/*******************************************************************************
*
* File wspace.c
*
* Copyright (C) 2010, 2011, 2013, 2015 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Workspace allocation and administration.
*
*   void alloc_wad(int n)
*     Allocates a workspace of n global sl3_alg_dble link fields.
*
*   sl3_alg_dble **reserve_wad(int n)
*     Reserves a new workspace of n global sl3_alg_dble link fields and
*     returns the array ad[0],..,ad[n-1] of the addresses of the fields
*     in the workspace. An error occurs if n<=0 or if space for too many
*     fields is requested.
*
*   int release_wad(void)
*     Releases the workspace of global sl3_alg_dble link fields that was
*     last reserved and returns the number of fields that were released.
*
*   int wad_size(void)
*     Returns the number of global sl3_alg_dble link fields that are
*     currently reserved.
*
*   void alloc_wcd(int n)
*     Allocates a workspace of n global sl3_alg_dble fields.
*
*   sl3_alg_dble **reserve_wcd(int n)
*     Reserves a new workspace of n global sl3_alg_dble fields and returns
*     the array cd[0],..,cd[n-1] of the addresses of the fields in the
*     workspace. An error occurs if n<=0 or if space for too many fields
*     is requested.
*
*   int release_wcd(void)
*     Releases the workspace of global sl3_alg_dble fields that was last
*     reserved and returns the number of fields that were released.
*
*   int wcd_size(void)
*     Returns the number of global sl3_alg_dble fields that are currently
*     reserved.
*
*   void alloc_wsd(int n)
*     Allocates a workspace of n global spinor_dble fields.
*
*   spinor_dble **reserve_wsd(int n)
*     Reserves a new workspace of n global spinor_dble fields and returns
*     the array sd[0],..,sd[n-1] of the addresses of the fields in the
*     workspace. An error occurs if n<=0 or if space for too many fields
*     is requested.
*
*   int release_wsd(void)
*     Releases the workspace of global spinor_dble fields that was last
*     reserved and returns the number of fields that were released.
*
*   int wsd_size(void)
*     Returns the number of global spinor_dble fields that are currently
*     reserved.
*
*   void alloc_wud(int n)
*     Allocates a workspace of n su3_dble link fields.
*
*   su3_dble **reserve_wud(int n)
*     Reserves a new workspace of n global su3_dble link fields and returns
*     the array ud[0],..,ud[n-1] of the addresses of the fields in the
*     workspace. An error occurs if n<=0 or if space for too many fields is
*     requested.
*
*   int release_wud(void)
*     Releases the workspace of global su3_dble link fields that was last
*     reserved and returns the number of fields that are released.
*
*   int wud_size(void)
*     Returns the number of global su3_dble link fields that are currently
*     reserved.
*
* By definition a workspace is a set of global fields which is reserved for
* the current program. This module allows to assign and release the required
* workspaces dynamically in such a way that the privacy of the workspaces is
* guaranteed.
*
* Workspaces reserved by a program must be released before the program returns
* to the calling program. It is possible to reserve several workspaces of the
* same type in a given program. In this case, a corresponding number of calls
* to w*_release() is required to release all of them.
*
* The sizes of the allocated fields are
*
*  alloc_wad()    4*VOLUME+7*(BNDRY/4)+3 if cpr[0]=NPROC0-1
*                 4*VOLUME+7*(BNDRY/4)   otherwise
*
*  alloc_wcd()    NSPIN
*
*  alloc_wsd()    NSPIN
*
*  alloc_wud()    same as alloc_wad()
*
* (see include/global.h and main/README.global). The allocated fields thus
* include the additional memory space required by the communication modules
* adcom.c, cdcom.c, sdcom.c and udcom.c.
*
* Reserved fields are laid out in memory in a contiguous way, i.e. subsequent
* addresses are separated exactly by the size of the fields listed above. The
* reserved workspaces may therefore be interpreted differently as a single
* field of the chosen type with n components per point rather than n fields
* with 1 component per point.
*
* Workspaces can always be freed and reallocated by calling alloc_w*() if no
* fields in the workspace are in use. In particular, a workspace can be freed
* by calling alloc_w*(0). Except for the w*_size() programs, all programs in
* this module must be called simultaneously on all MPI processes.
*
*******************************************************************************/

#define WSPACE_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "sl3.h"
#include "utils.h"
#include "global.h"

static int nadt=0,iwad=0,nwadt=0,*nwad;
static sl3_alg_dble **wad0,**wad,ad0={{0.0,0.0},{0.0,0.0},{0.0,0.0},{0.0,0.0},
                                      {0.0,0.0},{0.0,0.0},{0.0,0.0},{0.0,0.0}};

static int ncdt=0,iwcd=0,nwcdt=0,*nwcd;
static sl3_alg_dble **wcd0,**wcd,cd0={{0.0,0.0},{0.0,0.0},{0.0,0.0},{0.0,0.0},
                                      {0.0,0.0},{0.0,0.0},{0.0,0.0},{0.0,0.0}};

static int nsdt=0,iwsd=0,nwsdt=0,*nwsd;
static spinor_dble **wsd0,**wsd,sd0={{{0.0,0.0},{0.0,0.0},{0.0,0.0}},
                                     {{0.0,0.0},{0.0,0.0},{0.0,0.0}},
                                     {{0.0,0.0},{0.0,0.0},{0.0,0.0}},
                                     {{0.0,0.0},{0.0,0.0},{0.0,0.0}}};

static int nudt=0,iwud=0,nwudt=0,*nwud;
static su3_dble **wud0,**wud,ud0={{0.0,0.0},{0.0,0.0},{0.0,0.0},
                                  {0.0,0.0},{0.0,0.0},{0.0,0.0},
                                  {0.0,0.0},{0.0,0.0},{0.0,0.0}};


static void check_parm(char *name,int n)
{
   int iprms[1];

   if (NPROC>1)
   {
      iprms[0]=n;
      MPI_Bcast(iprms,1,MPI_INT,0,MPI_COMM_WORLD);
      error(iprms[0]!=n,1,name,"Parameter n is not global");
   }
}


void alloc_wad(int n)
{
   int m,i;
   sl3_alg_dble *ad,*am;

   check_parm("alloc_wad [wspace.c]",n);

   if (n==nadt)
      return;

   error_root(n<0,1,"alloc_wad [wspace.c]","Parameter n must be non-negative");
   error_root(nwadt!=0,1,"alloc_wad [wspace.c]","Workspace is in use");

   if (nadt>0)
   {
      free(nwad);
      afree(wad0[0]);
      free(wad0);
      nwad=NULL;
      wad0=NULL;
      wad=NULL;
   }

   nadt=n;
   iwad=0;
   nwadt=0;

   if (nadt>0)
   {
      nwad=malloc(nadt*sizeof(*nwad));
      wad0=malloc(2*nadt*sizeof(*wad0));
      wad=wad0+nadt;

      error((nwad==NULL)||(wad0==NULL),1,"alloc_wad [wspace.c]",
            "Unable to allocate index arrays");

      m=4*VOLUME+7*(BNDRY/4);
      if (cpr[0]==(NPROC0-1))
         m+=3;
      wad0[0]=amalloc(nadt*m*sizeof(**wad0),ALIGN);

      error(wad0[0]==NULL,1,"alloc_wad [wspace.c]",
            "Unable to allocate workspace");

      for (i=0;i<nadt;i++)
      {
         if (i>0)
            wad0[i]=wad0[i-1]+m;

         nwad[i]=0;
         wad[i]=NULL;
      }

      ad=wad0[0];
      am=ad+nadt*m;

      for (;ad<am;ad++)
         (*ad)=ad0;
   }
}


sl3_alg_dble **reserve_wad(int n)
{
   int i,ia;

   check_parm("reserve_wad [wspace.c]",n);

   if ((n>0)&&((nwadt+n)<=nadt))
   {
      ia=nwadt;
      nwad[iwad]=n;
      nwadt+=n;
      iwad+=1;

      for (i=ia;i<(ia+n);i++)
         wad[i]=wad0[i];

      return wad+ia;
   }
   else
   {
      error(n<=0,1,"reserve_wad [wspace.c]",
            "Number of requested fields must be positive");
      error((nwadt+n)>nadt,1,"reserve_wad [wspace.c]",
            "Requested too many fields (tot=%d,use=%d,req=%d)",nadt,nwadt,n);
      return NULL;
   }
}


int release_wad(void)
{
   int n,i;

   error(iwad<=0,1,"release_wad [wspace.c]",
         "No workspace is currently reserved");

   iwad-=1;
   n=nwad[iwad];
   nwadt-=n;
   nwad[iwad]=0;

   for (i=nwadt;i<(nwadt+n);i++)
      wad[i]=NULL;

   return n;
}


int wad_size(void)
{
   return nwadt;
}


void alloc_wcd(int n)
{
   int m,i;
   sl3_alg_dble *cd,*cm;

   check_parm("alloc_wcd [wspace.c]",n);

   if (n==ncdt)
      return;

   error_root(n<0,1,"alloc_wcd [wspace.c]","Parameter n must be non-negative");
   error_root(nwcdt!=0,1,"alloc_wcd [wspace.c]","Workspace is in use");

   if (ncdt>0)
   {
      free(nwcd);
      afree(wcd0[0]);
      free(wcd0);
      nwcd=NULL;
      wcd0=NULL;
      wcd=NULL;
   }

   ncdt=n;
   iwcd=0;
   nwcdt=0;

   if (ncdt>0)
   {
      nwcd=malloc(ncdt*sizeof(*nwcd));
      wcd0=malloc(2*ncdt*sizeof(*wcd0));
      wcd=wcd0+ncdt;

      error((nwcd==NULL)||(wcd0==NULL),1,"alloc_wcd [wspace.c]",
            "Unable to allocate index arrays");

      m=NSPIN;
      wcd0[0]=amalloc(ncdt*m*sizeof(**wcd0),ALIGN);

      error(wcd0[0]==NULL,1,"alloc_wcd [wspace.c]",
            "Unable to allocate workspace");

      for (i=0;i<ncdt;i++)
      {
         if (i>0)
            wcd0[i]=wcd0[i-1]+m;

         nwcd[i]=0;
         wcd[i]=NULL;
      }

      cd=wcd0[0];
      cm=cd+ncdt*m;

      for (;cd<cm;cd++)
         (*cd)=cd0;
   }
}


sl3_alg_dble **reserve_wcd(int n)
{
   int i,ia;

   check_parm("reserve_wcd [wspace.c]",n);

   if ((n>0)&&((nwcdt+n)<=ncdt))
   {
      ia=nwcdt;
      nwcd[iwcd]=n;
      nwcdt+=n;
      iwcd+=1;

      for (i=ia;i<(ia+n);i++)
         wcd[i]=wcd0[i];

      return wcd+ia;
   }
   else
   {
      error(n<=0,1,"reserve_wcd [wspace.c]",
            "Number of requested fields must be positive");
      error((nwcdt+n)>ncdt,1,"reserve_wcd [wspace.c]",
            "Requested too many fields (tot=%d,use=%d,req=%d)",ncdt,nwcdt,n);
      return NULL;
   }
}


int release_wcd(void)
{
   int n,i;

   error(iwcd<=0,1,"release_wcd [wspace.c]",
         "No workspace is currently reserved");

   iwcd-=1;
   n=nwcd[iwcd];
   nwcdt-=n;
   nwcd[iwcd]=0;

   for (i=nwcdt;i<(nwcdt+n);i++)
      wcd[i]=NULL;

   return n;
}


int wcd_size(void)
{
   return nwcdt;
}


void alloc_wsd(int n)
{
   int m,i;
   spinor_dble *sd,*sm;

   check_parm("alloc_wsd [wspace.c]",n);

   if (n==nsdt)
      return;

   error_root(n<0,1,"alloc_wsd [wspace.c]","Parameter n must be non-negative");
   error_root(nwsdt!=0,1,"alloc_wsd [wspace.c]","Workspace is in use");

   if (nsdt>0)
   {
      free(nwsd);
      afree(wsd0[0]);
      free(wsd0);
      nwsd=NULL;
      wsd0=NULL;
      wsd=NULL;
   }

   nsdt=n;
   iwsd=0;
   nwsdt=0;

   if (nsdt>0)
   {
      nwsd=malloc(nsdt*sizeof(*nwsd));
      wsd0=malloc(2*nsdt*sizeof(*wsd0));
      wsd=wsd0+nsdt;

      error((nwsd==NULL)||(wsd0==NULL),1,"alloc_wsd [wspace.c]",
            "Unable to allocate index arrays");

      m=NSPIN;
      wsd0[0]=amalloc(nsdt*m*sizeof(**wsd0),ALIGN);

      error(wsd0[0]==NULL,1,"alloc_wsd [wspace.c]",
            "Unable to allocate workspace");

      for (i=0;i<nsdt;i++)
      {
         if (i>0)
            wsd0[i]=wsd0[i-1]+m;

         nwsd[i]=0;
         wsd[i]=NULL;
      }

      sd=wsd0[0];
      sm=sd+nsdt*m;

      for (;sd<sm;sd++)
         (*sd)=sd0;
   }
}


spinor_dble **reserve_wsd(int n)
{
   int i,ia;

   check_parm("reserve_wsd [wspace.c]",n);

   if ((n>0)&&((nwsdt+n)<=nsdt))
   {
      ia=nwsdt;
      nwsd[iwsd]=n;
      nwsdt+=n;
      iwsd+=1;

      for (i=ia;i<(ia+n);i++)
         wsd[i]=wsd0[i];

      return wsd+ia;
   }
   else
   {
      error(n<=0,1,"reserve_wsd [wspace.c]",
            "Number of requested fields must be positive");
      error((nwsdt+n)>nsdt,1,"reserve_wsd [wspace.c]",
            "Requested too many fields (tot=%d,use=%d,req=%d)",nsdt,nwsdt,n);
      return NULL;
   }
}


int release_wsd(void)
{
   int n,i;

   error(iwsd<=0,1,"release_wsd [wspace.c]",
         "No workspace is currently reserved");

   iwsd-=1;
   n=nwsd[iwsd];
   nwsdt-=n;
   nwsd[iwsd]=0;

   for (i=nwsdt;i<(nwsdt+n);i++)
      wsd[i]=NULL;

   return n;
}


int wsd_size(void)
{
   return nwsdt;
}


void alloc_wud(int n)
{
   int m,i;
   su3_dble *ud,*um;

   check_parm("alloc_wud [wspace.c]",n);

   if (n==nudt)
      return;

   error_root(n<0,1,"alloc_wud [wspace.c]","Parameter n must be non-negative");
   error_root(nwudt!=0,1,"alloc_wud [wspace.c]","Workspace is in use");

   if (nudt>0)
   {
      free(nwud);
      afree(wud0[0]);
      free(wud0);
      nwud=NULL;
      wud0=NULL;
      wud=NULL;
   }

   nudt=n;
   iwud=0;
   nwudt=0;

   if (nudt>0)
   {
      nwud=malloc(nudt*sizeof(*nwud));
      wud0=malloc(2*nudt*sizeof(*wud0));
      wud=wud0+nudt;

      error((nwud==NULL)||(wud0==NULL),1,"alloc_wud [wspace.c]",
            "Unable to allocate index arrays");

      m=4*VOLUME+7*(BNDRY/4);
      if (cpr[0]==(NPROC0-1))
         m+=3;
      wud0[0]=amalloc(nudt*m*sizeof(**wud0),ALIGN);

      error(wud0[0]==NULL,1,"alloc_wud [wspace.c]",
            "Unable to allocate workspace");

      for (i=0;i<nudt;i++)
      {
         if (i>0)
            wud0[i]=wud0[i-1]+m;

         nwud[i]=0;
         wud[i]=NULL;
      }

      ud=wud0[0];
      um=ud+nudt*m;

      for (;ud<um;ud++)
         (*ud)=ud0;
   }
}


su3_dble **reserve_wud(int n)
{
   int i,ia;

   check_parm("reserve_wud [wspace.c]",n);

   if ((n>0)&&((nwudt+n)<=nudt))
   {
      ia=nwudt;
      nwud[iwud]=n;
      nwudt+=n;
      iwud+=1;

      for (i=ia;i<(ia+n);i++)
         wud[i]=wud0[i];

      return wud+ia;
   }
   else
   {
      error(n<=0,1,"reserve_wud [wspace.c]",
            "Number of requested fields must be positive");
      error((nwudt+n)>nudt,1,"reserve_wud [wspace.c]",
            "Requested too many fields (tot=%d,use=%d,req=%d)",nudt,nwudt,n);
      return NULL;
   }
}


int release_wud(void)
{
   int n,i;

   error(iwud<=0,1,"release_wud [wspace.c]",
         "No workspace is currently reserved");

   iwud-=1;
   n=nwud[iwud];
   nwudt-=n;
   nwud[iwud]=0;

   for (i=nwudt;i<(nwudt+n);i++)
      wud[i]=NULL;

   return n;
}


int wud_size(void)
{
   return nwudt;
}
