
/*******************************************************************************
*
* File AGn.c
*
* Copyright (C) 2015, 2017 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Computation of the vertex functions A_G^n.
*
*   void AGn(int n,sl3_alg_dble **asrc,sl3_alg_dble *vxfld)
*     Assigns the vertex function A_G^n to the field vxfld for given
*     source fields asrc[0],..,asrc[n-2]. The latter are assumed to
*     satisfy the boundary conditions specified in the parameter data
*     base.
*
* See the notes
*
*  M. Luescher: "Lattice QCD vertices in ISPT", file doc/vxfcts.pdf,
*
* for the definition of the vertex functions A_G^n and a description of
* the computational strategy employed.
*
* Gauge potentials are locally stored as described in main/README.global
* and aflds/README.aflds. Both input and output fields must be of this
* type. The output field may not overlap with any of the input fields.
*
* The parameter n must be in the range 3<=n<=NMAX, where NMAX is defined
* in vxfcts.h.
*
* The program in this module acts globally and must be called on all MPI
* processes simultaneously.
*
*******************************************************************************/

#define AGN_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "sl3.h"
#include "utils.h"
#include "lattice.h"
#include "su3fcts.h"
#include "linalg.h"
#include "aflds.h"
#include "vxfcts.h"
#include "global.h"

static int ns=0;
static double fact[NMAX+1],ifact[NMAX+1];
static su3_dble *Ak,*eA[4],*dPY[4],*dQM[4];
static su3_dble A1s[4] ALIGNED16;
static su3_dble A2s[4] ALIGNED16;
static su3_dble Xs[6] ALIGNED16;


static void alloc_arrays(int n)
{
   int nd,i;

   nd=1<<(n-1);

   if (n>ns)
   {
      if (ns==0)
      {
         fact[0]=1.0;
         ifact[0]=1.0;

         for (i=1;i<=NMAX;i++)
         {
            fact[i]=(double)(i)*fact[i-1];
            ifact[i]=1.0/fact[i];
         }
      }
      else
         afree(Ak);

      Ak=amalloc((n-1+12*nd)*sizeof(*Ak),ALIGN);
      error(Ak==NULL,1,"alloc_arrays [AGn.c]",
            "Unable to allocate auxiliary arrays");

      cm3x3_zero(n-1+12*nd,Ak);
      ns=n;
   }

   eA[0]=Ak+n-1;
   dPY[0]=eA[0]+4*nd;
   dQM[0]=dPY[0]+4*nd;

   for (i=1;i<4;i++)
   {
      eA[i]=eA[i-1]+nd;
      dPY[i]=dPY[i-1]+nd;
      dQM[i]=dQM[i-1]+nd;
   }
}


static int not_yet(int i,sl3_alg_dble **asrc)
{
   int j;

   for (j=0;j<i;j++)
   {
      if (asrc[j]==asrc[i])
         return 0;
   }

   return 1;
}


static void copy_bnd(int n,sl3_alg_dble **asrc)
{
   int i;

   for (i=0;i<(n-1);i++)
   {
      if (not_yet(i,asrc))
         copy_bnd_ad(0,asrc[i]);
   }
}


void AGn3(int *ip,sl3_alg_dble **asrc)
{
   int k,l;
   su3_dble *X;

   for (k=0;k<4;k++)
   {
      sl3_to_su3(asrc[0]+ip[k],A1s+k);
      sl3_to_su3(asrc[1]+ip[k],A2s+k);
   }

   X=Xs;

   for (k=1;k<4;k++)
   {
      for (l=0;l<k;l++)
      {
         su3xsu3(A1s+k,A2s+l,X);
         cm3x3_mul_sub(A2s+l,A1s+k,X);
         cm3x3_mul_add(A2s+k,A1s+l,X);
         cm3x3_mul_sub(A1s+l,A2s+k,X);
         X+=1;
      }
   }

   A1s[0]=Xs[5];
   cm3x3_add(Xs+2,A1s);
   cm3x3_add(Xs+4,A1s);

   A1s[1]=Xs[5];
   cm3x3_sub(Xs+1,A1s+1);
   cm3x3_sub(Xs+3,A1s+1);

   A1s[2]=Xs[0];
   cm3x3_sub(Xs+3,A1s+2);
   cm3x3_sub(Xs+4,A1s+2);

   A1s[3]=Xs[0];
   cm3x3_add(Xs+1,A1s+3);
   cm3x3_add(Xs+2,A1s+3);
}


static void set_eA(int n,int *ip,sl3_alg_dble **asrc)
{
   int k,i;

   for (k=0;k<4;k++)
   {
      for (i=0;i<(n-1);i++)
         sl3_to_su3(asrc[i]+ip[k],Ak+i);

      sym_prod(0,n-1,Ak,eA[k]);
   }
}


static void set_dPY(int n)
{
   int p,ip,is,ipmx,ismx;
   int ius,iur;
   double sgn;
   su3_dble *X[4];

   sgn=-1.0;
   ipmx=1<<(n-1);
   X[0]=dPY[0];
   X[1]=dPY[1];
   X[2]=dPY[2];
   X[3]=dPY[3];

   for (ip=0;ip<ipmx;ip++)
   {
      X[0][0]=eA[1][ip];
      X[1][0]=eA[0][ip];
      X[2][0]=eA[3][ip];
      X[3][0]=eA[2][ip];

      if (ip>0)
      {
         cm3x3_add(eA[0]+ip,X[0]);
         cm3x3_add(eA[1]+ip,X[1]);
         cm3x3_add(eA[2]+ip,X[2]);
         cm3x3_add(eA[3]+ip,X[3]);
      }

      p=bcount(n-1,ip);
      ismx=1<<p;

      for (is=1;is<(ismx-1);is++)
      {
         ius=bsubset(n-1,is,ip);
         iur=ip^ius;

         cm3x3_mul_add(eA[0]+ius,eA[1]+iur,X[0]);
         cm3x3_mul_add(eA[1]+ius,eA[0]+iur,X[1]);
         cm3x3_mul_add(eA[2]+ius,eA[3]+iur,X[2]);
         cm3x3_mul_add(eA[3]+ius,eA[2]+iur,X[3]);
      }

      if (p&0x1)
      {
         cm3x3_mulr(&sgn,X[0],X[0]);
         cm3x3_mulr(&sgn,X[1],X[1]);
         cm3x3_mulr(&sgn,X[2],X[2]);
         cm3x3_mulr(&sgn,X[3],X[3]);
      }

      X[0]+=1;
      X[1]+=1;
      X[2]+=1;
      X[3]+=1;
   }
}


static void set_dQM(int n)
{
   int p,ip,is,ipmx,ismx;
   int ius,iur;
   double two,*r;
   su3_dble *X[4];

   two=2.0;
   ipmx=1<<(n-1);
   X[0]=dQM[0];
   X[1]=dQM[1];
   X[2]=dQM[2];
   X[3]=dQM[3];

   for (ip=0;ip<ipmx;ip++)
   {
      X[0][0]=dPY[3][ip];
      X[1][0]=dPY[3][ip];
      X[2][0]=dPY[1][ip];
      X[3][0]=dPY[1][ip];

      if (n&0x1)
      {
         cm3x3_sub(dPY[2]+ip,X[0]);
         cm3x3_sub(dPY[2]+ip,X[1]);
         cm3x3_sub(dPY[0]+ip,X[2]);
         cm3x3_sub(dPY[0]+ip,X[3]);
      }
      else
      {
         cm3x3_add(dPY[2]+ip,X[0]);
         cm3x3_add(dPY[2]+ip,X[1]);
         cm3x3_add(dPY[0]+ip,X[2]);
         cm3x3_add(dPY[0]+ip,X[3]);

         if (ip>0)
         {
            cm3x3_mulr_add(&two,eA[1]+ip,X[0]);
            cm3x3_mulr_add(&two,eA[0]+ip,X[1]);
            cm3x3_mulr_add(&two,eA[3]+ip,X[2]);
            cm3x3_mulr_add(&two,eA[2]+ip,X[3]);
         }
      }

      p=bcount(n-1,ip);
      ismx=1<<p;

      for (is=1;is<(ismx-1);is++)
      {
         ius=bsubset(n-1,is,ip);
         iur=ip^ius;

         cm3x3_mul_add(eA[1]+iur,dPY[3]+ius,X[0]);
         cm3x3_mul_add(dPY[3]+ius,eA[0]+iur,X[1]);
         cm3x3_mul_add(eA[3]+iur,dPY[1]+ius,X[2]);
         cm3x3_mul_add(dPY[1]+ius,eA[2]+iur,X[3]);

         if (n&0x1)
         {
            cm3x3_mul_sub(dPY[2]+ius,eA[1]+iur,X[0]);
            cm3x3_mul_sub(eA[0]+iur,dPY[2]+ius,X[1]);
            cm3x3_mul_sub(dPY[0]+ius,eA[3]+iur,X[2]);
            cm3x3_mul_sub(eA[2]+iur,dPY[0]+ius,X[3]);
         }
         else
         {
            cm3x3_mul_add(dPY[2]+ius,eA[1]+iur,X[0]);
            cm3x3_mul_add(eA[0]+iur,dPY[2]+ius,X[1]);
            cm3x3_mul_add(dPY[0]+ius,eA[3]+iur,X[2]);
            cm3x3_mul_add(eA[2]+iur,dPY[0]+ius,X[3]);
         }
      }

      if (p<(n-1))
      {
         r=ifact+n-p;
         cm3x3_mulr(r,X[0],X[0]);
         cm3x3_mulr(r,X[1],X[1]);
         cm3x3_mulr(r,X[2],X[2]);
         cm3x3_mulr(r,X[3],X[3]);
      }

      X[0]+=1;
      X[1]+=1;
      X[2]+=1;
      X[3]+=1;
   }
}


static void reset_eA(int n)
{
   int p,ip,ipmx;
   double *r;
   su3_dble *X[4];

   ipmx=1<<(n-1);
   X[0]=eA[0];
   X[1]=eA[1];
   X[2]=eA[2];
   X[3]=eA[3];

   for (ip=0;ip<ipmx;ip++)
   {
      p=bcount(n-1,ip);

      if (p>1)
      {
         r=fact+p;
         cm3x3_mulr(r,X[0],X[0]);
         cm3x3_mulr(r,X[1],X[1]);
         cm3x3_mulr(r,X[2],X[2]);
         cm3x3_mulr(r,X[3],X[3]);
      }

      X[0]+=1;
      X[1]+=1;
      X[2]+=1;
      X[3]+=1;
   }
}


static void inner_prod(int n)
{
   int p,ip,is,ipmx,ismx;
   int ius,iur;
   su3_dble *X[4];

   ipmx=1<<(n-1);
   X[0]=dPY[0];
   X[1]=dPY[1];
   X[2]=dPY[2];
   X[3]=dPY[3];

   for (ip=0;ip<ipmx;ip++)
   {
      X[0][0]=dQM[0][ip];
      X[1][0]=dQM[1][ip];
      X[2][0]=dQM[2][ip];
      X[3][0]=dQM[3][ip];

      p=bcount(n-1,ip);
      ismx=1<<p;

      for (is=1;is<ismx;is++)
      {
         ius=bsubset(n-1,is,ip);
         iur=ip^ius;

         cm3x3_mul_add(eA[0]+ius,dQM[0]+iur,X[0]);
         cm3x3_mul_add(eA[1]+ius,dQM[1]+iur,X[1]);
         cm3x3_mul_add(eA[2]+ius,dQM[2]+iur,X[2]);
         cm3x3_mul_add(eA[3]+ius,dQM[3]+iur,X[3]);
      }

      X[0]+=1;
      X[1]+=1;
      X[2]+=1;
      X[3]+=1;
   }
}


static void outer_prod(int n)
{
   int ir,is,ipmx;

   ipmx=1<<(n-1);
   A1s[0]=dPY[0][ipmx-1];
   A1s[1]=dPY[1][ipmx-1];
   A1s[2]=dPY[2][ipmx-1];
   A1s[3]=dPY[3][ipmx-1];

   for (ir=1;ir<ipmx;ir++)
   {
      is=(ipmx-1)^ir;

      cm3x3_mul_add(dPY[0]+is,eA[0]+ir,A1s);
      cm3x3_mul_add(dPY[1]+is,eA[1]+ir,A1s+1);
      cm3x3_mul_add(dPY[2]+is,eA[2]+ir,A1s+2);
      cm3x3_mul_add(dPY[3]+is,eA[3]+ir,A1s+3);
   }
}


void AGn(int n,sl3_alg_dble **asrc,sl3_alg_dble *vxfld)
{
   int iprms[1];
   int ix,t,np,ip[4];
   double wc;

   error_root((n<3)||(n>NMAX),1,"AGn [AGn.c]",
              "Parameter n is out of range");

   if (NPROC>1)
   {
      iprms[0]=n;
      MPI_Bcast(iprms,1,MPI_INT,0,MPI_COMM_WORLD);
      error(iprms[0]!=n,1,"AGn [AGn.c]","Parameter n is not global");
   }

   alloc_arrays(n);
   copy_bnd(n,asrc);

   if (cpr[0]==(NPROC0-1))
      set_calg2zero(4*VOLUME+7*(BNDRY/4)+3,vxfld);
   else
      set_calg2zero(4*VOLUME+7*(BNDRY/4),vxfld);

   for (ix=0;ix<VOLUME;ix++)
   {
      t=global_time(ix);

      for (np=0;np<6;np++)
      {
         plaq_aidx(np,ix,ip);

         if (n==3)
            AGn3(ip,asrc);
         else
         {
            set_eA(n,ip,asrc);
            set_dPY(n);
            set_dQM(n);
            reset_eA(n);
            inner_prod(n);
            outer_prod(n);
         }

         if ((t==0)&&(np>=3))
            wc=-0.25;
         else
            wc=-0.5;

         su3_mulr_add_sl3(&wc,A1s  ,vxfld+ip[0]);
         su3_mulr_add_sl3(&wc,A1s+1,vxfld+ip[1]);
         su3_mulr_add_sl3(&wc,A1s+2,vxfld+ip[2]);
         su3_mulr_add_sl3(&wc,A1s+3,vxfld+ip[3]);
      }
   }

   add_bnd_ad(0,vxfld);
}
