
/*******************************************************************************
*
* File Agf.c
*
* Copyright (C) 2015 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Computation of the vertex functions A_gf.
*
*   void Agf(sl3_alg_dble *asrc,sl3_alg_dble *vxfld)
*     Assigns the vertex function A_gf to the field vxfld for given
*     source field asrc. The boundary conditions are determined from
*     the parameter data base and it is taken for granted that the
*     source field satisfies the boundary conditions.
*
* See the notes
*
*  M. Luescher: "Lattice QCD vertices in ISPT", file doc/vxfcts.pdf,
*
* for the definition of the vertex function A_gf and a description of
* the computational strategy employed.
*
* Gauge potentials are locally stored as described in main/README.global
* and aflds/README.aflds. Both input and output fields must be of this
* type. The input field is unchanged unless it coincides with the output
* field (which is permissible).
*
* The program in this module acts globally and must be called on all MPI
* processes simultaneously. It requires a workspace of 1 ghost field.
*
*******************************************************************************/

#define AGF_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "sl3.h"
#include "utils.h"
#include "pmdb.h"
#include "lattice.h"
#include "su3fcts.h"
#include "linalg.h"
#include "aflds.h"
#include "cflds.h"
#include "vxfcts.h"
#include "global.h"

#define N0 (NPROC0*L0)
#define N1 (NPROC1*L1)
#define N2 (NPROC2*L2)
#define N3 (NPROC3*L3)

static int init=0,idsm;
static sl3_alg_dble F0 ALIGNED16;


static void set_F0(sl3_alg_dble *asrc)
{
   int ix;
   double r;

   if (init==0)
   {
      idsm=init_hsum(16);
      init=1;
   }

   reset_hsum(idsm);

   for (ix=(VOLUME/2);ix<VOLUME;ix++)
   {
      add_to_hsum(idsm,(double*)(asrc));
      asrc+=1;
      add_to_hsum(idsm,(double*)(asrc));
      asrc+=7;
   }

   global_hsum(idsm,(double*)(&F0));

   r=1.0/((double)(N0*N0*N0)*(double)(N1*N2*N3));
   sl3_mulr(&r,&F0,&F0);
}


void Agf(sl3_alg_dble *asrc,sl3_alg_dble *vxfld)
{
   int ix,iy,t,ifc,mu;
   double half;
   sl3_alg_dble **wcd,*F;

   wcd=reserve_wcd(1);
   F=wcd[0];
   divad(asrc,F);
   cpcd_int_bnd(0,F);
   half=0.5;

   if (bc_type()==1)
   {
      set_F0(asrc);

      for (ix=(VOLUME/2);ix<VOLUME;ix++)
      {
         t=global_time(ix);

         for (ifc=0;ifc<8;ifc++)
         {
            mu=ifc/2;

            if (ifc&0x1)
            {
               if (t!=0)
                  (*vxfld)=F[ix];
               else
                  sl3_zero(vxfld);

               if (((mu==0)&&(t!=1))||((mu>0)&&(t!=0)))
               {
                  iy=idn[ix][mu];
                  sl3_sub(F+iy,vxfld);
               }
            }
            else
            {
               if (((mu==0)&&(t!=(N0-1)))||((mu>0)&&(t!=0)))
               {
                  iy=iup[ix][mu];
                  (*vxfld)=F[iy];
               }
               else
                  sl3_zero(vxfld);

               if (t!=0)
                  sl3_sub(F+ix,vxfld);
            }

            if (mu==0)
               sl3_sub(&F0,vxfld);

            vxfld+=1;
         }
      }
   }
   else
   {
      for (ix=(VOLUME/2);ix<VOLUME;ix++)
      {
         t=global_time(ix);

         for (ifc=0;ifc<8;ifc++)
         {
            mu=ifc/2;

            if (ifc&0x1)
            {
               if ((t!=0)||(mu>0))
                  (*vxfld)=F[ix];
               else
                  sl3_zero(vxfld);

               iy=idn[ix][mu];
               sl3_sub(F+iy,vxfld);
            }
            else
            {
               if ((t!=(N0-1))||(mu>0))
               {
                  iy=iup[ix][mu];
                  (*vxfld)=F[iy];
               }
               else
                  sl3_zero(vxfld);

               sl3_sub(F+ix,vxfld);
            }

            if ((t==0)&&(mu>0))
               sl3_mulr(&half,vxfld,vxfld);

            vxfld+=1;
         }
      }
   }

   if (cpr[0]==(NPROC0-1))
      set_calg2zero(3,vxfld+7*(BNDRY/4));

   release_wcd();
}
